from __future__ import annotations

from typing import TYPE_CHECKING

from sdc11073.namespaces import PrefixesEnum

from .porttypebase import (
    DPWSPortTypeBase,
    WSDLMessageDescription,
    WSDLOperationBinding,
    mk_wsdl_one_way_operation,
    msg_prefix,
)

if TYPE_CHECKING:
    from sdc11073 import xml_utils
    from sdc11073.mdib.descriptorcontainers import AbstractDescriptorContainer
    from sdc11073.mdib.mdibbase import MdibVersionGroup
    from sdc11073.mdib.statecontainers import AbstractStateContainer


class DescriptionEventService(DPWSPortTypeBase):
    port_type_name = PrefixesEnum.SDC.tag('DescriptionEventService')
    WSDLMessageDescriptions = (
        WSDLMessageDescription('DescriptionModificationReport',
                               (f'{msg_prefix}:DescriptionModificationReport',)),
    )
    WSDLOperationBindings = (WSDLOperationBinding('DescriptionModificationReport', None, 'literal'),
                             )

    def add_wsdl_port_type(self, parent_node):
        port_type = self._mk_port_type_node(parent_node, True)
        mk_wsdl_one_way_operation(port_type, operation_name='DescriptionModificationReport')

    def send_descriptor_updates(self, updated: list[AbstractDescriptorContainer],
                                created: list[AbstractDescriptorContainer],
                                deleted: list[AbstractDescriptorContainer],
                                updated_states: list[AbstractStateContainer],
                                mdib_version_group: MdibVersionGroup):
        subscription_mgr = self.hosting_service.subscriptions_manager
        action = self._sdc_definitions.Actions.DescriptionModificationReport
        body_node = self.mk_description_modification_report_body(
            mdib_version_group, updated, created, deleted, updated_states)
        self._logger.debug('sending DescriptionModificationReport upd={} crt={} del={}', updated, created, deleted)
        subscription_mgr.send_to_subscribers(body_node, action.value, mdib_version_group)

    def mk_description_modification_report_body(self,
                                                mdib_version_group: MdibVersionGroup,
                                                updated: list[AbstractDescriptorContainer],
                                                created: list[AbstractDescriptorContainer],
                                                deleted: list[AbstractDescriptorContainer],
                                                updated_states: list[AbstractStateContainer]) -> xml_utils.LxmlElement:
        # This method creates one ReportPart for every descriptor.
        # An optimization is possible by grouping all descriptors with the same parent handle into one ReportPart.
        # This is not implemented, and I think it is not needed.
        data_model = self._sdc_definitions.data_model
        report = data_model.msg_types.DescriptionModificationReport()
        report.set_mdib_version_group(mdib_version_group)
        DescriptionModificationType = data_model.msg_types.DescriptionModificationType

        for descriptors, modification_type in ((updated, DescriptionModificationType.UPDATE),
                                               (created, DescriptionModificationType.CREATE),
                                               (deleted, DescriptionModificationType.DELETE)):
            for descriptor in descriptors:
                # one report part for every descriptor,
                report_part = report.add_report_part()
                report_part.ModificationType = modification_type
                report_part.ParentDescriptor = descriptor.parent_handle
                report_part.SourceMds = descriptor.source_mds
                report_part.Descriptor.append(descriptor)
                states = [s for s in updated_states if s.DescriptorHandle == descriptor.Handle]
                report_part.State.extend(states)

        nsh = data_model.ns_helper
        ns_map = nsh.partial_map(nsh.MSG, nsh.PM)
        return report.as_etree_node(data_model.msg_names.DescriptionModificationReport, ns_map)



from __future__ import annotations

import queue
import threading
import traceback
from typing import TYPE_CHECKING

from sdc11073.dispatch import RequestData, RequestDispatcher
from sdc11073.exceptions import InvalidActionError
from sdc11073.pysoap.msgfactory import CreatedMessage
from sdc11073.pysoap.soapenvelope import Fault, faultcodeEnum

if TYPE_CHECKING:
    from .manipulator import RequestManipulatorProtocol


class EmptyResponse(CreatedMessage):
    """EmptyResponse is a response with no content."""

    def __init__(self):
        super().__init__(None, None)

    def serialize(self, pretty: bool = False,  # noqa: ARG002
                  request_manipulator: RequestManipulatorProtocol | None = None,  # noqa: ARG002
                  validate: bool = True) -> bytes:  # noqa: ARG002
        """Return bytes of len 0."""
        return b''


class DispatchKeyRegistryDeferred(RequestDispatcher):
    """A middleware that splits request processing into two parts.

    It writes the request to a queue and returns immediately. A worker thread is responsible for the further handling.
    This allows a faster response.
    """

    def __init__(self, log_prefix: str):
        super().__init__(log_prefix)
        self._queue = queue.Queue(1000)
        self._worker = threading.Thread(target=self._read_queue)
        self._worker.daemon = True
        self._worker.start()

    def on_post(self, request_data: RequestData) -> CreatedMessage:
        """See documentation in RequestHandlerProtocol."""
        action = request_data.message_data.action
        func = self._get_post_handler(request_data)
        if func is None:
            fault = Fault()
            fault.Code.Value = faultcodeEnum.SENDER
            fault.add_reason_text(f'invalid action {action}')

            raise InvalidActionError(fault)
        self._queue.put((func, request_data, action))
        return EmptyResponse()

    def _read_queue(self):
        while True:
            func, request, action = self._queue.get()
            try:
                func(request)
            except Exception:  # noqa: BLE001
                # catch all to keep thread alive
                self._logger.error('method {} for action "{}" failed:{}',  # noqa: PLE1205
                                   func.__name__, action, traceback.format_exc())



from sdc11073.observableproperties.observables import ObservableProperty, bind, strongbind, unbind
from sdc11073.observableproperties.valuecollector import (
    CancelledError,
    CollectTimeoutError,
    SingleValueCollector,
    ValuesCollector,
)

__all__ = [
    'bind',
    'strongbind',
    'unbind',
    'ObservableProperty',
    'SingleValueCollector',
    'ValuesCollector',
    'CancelledError',
    'CollectTimeoutError',
]



"""The module implements a waveform provider."""
from __future__ import annotations

import time
import traceback
from decimal import Context
from threading import Event, Thread
from typing import TYPE_CHECKING, Any, Protocol

from sdc11073 import loghelper
from sdc11073.intervaltimer import IntervalTimer

from . import waveforms
from .realtimesamples import Annotator, RtSampleArray

if TYPE_CHECKING:
    from sdc11073.definitions_base import AbstractDataModel
    from sdc11073.mdib.entityprotocol import EntityProtocol
    from sdc11073.mdib.mdibprotocol import ProviderMdibProtocol
    from sdc11073.mdib.statecontainers import RealTimeSampleArrayMetricStateContainer
    from sdc11073.xml_types.pm_types import ComponentActivation

    from .realtimesamples import AnnotatorProtocol


class WaveformGeneratorProtocol(Protocol):
    """A waveform generator creates an infinite sequence of float values."""

    def __init__(self, min_value: float, max_value: float, waveform_period: float, sample_period: float):
        ...

    def next_samples(self, count: int) -> list[float]:
        """Get next values from generator."""
        ...

    sample_period: float


class _SampleArrayGenerator:
    """Wraps a waveform generator and makes RtSampleArray objects."""

    def __init__(self, model: AbstractDataModel,
                 descriptor_handle: str,
                 generator: WaveformGeneratorProtocol):
        self._model = model
        self._descriptor_handle = descriptor_handle
        self._last_timestamp = None
        self._activation_state = model.pm_types.ComponentActivation.ON
        self._generator = generator
        self.current_rt_sample_array = None

    def set_activation_state(self, component_activation_state: ComponentActivation):
        """Set activation state of generator.

        If component_activation_state is not "ON", the generator will not generate values.
        :param component_activation_state: one of pmtypes.ComponentActivation values.
        """
        self._activation_state = component_activation_state
        if component_activation_state == self._model.pm_types.ComponentActivation.ON:
            self._last_timestamp = time.time()

    def get_next_sample_array(self) -> RtSampleArray:
        """Read sample values from waveform generator and calculate determination time.

        If activation state is not 'On', the returned RtSampleArray contains no samples.
        """
        if self._activation_state != self._model.pm_types.ComponentActivation.ON:
            self.current_rt_sample_array = RtSampleArray(
                self._model, None, self._generator.sample_period, [], self._activation_state)
        else:
            now = time.time()
            observation_time = self._last_timestamp or now
            samples_count = int((now - observation_time) / self._generator.sample_period)
            samples = self._generator.next_samples(samples_count)
            self._last_timestamp = observation_time + self._generator.sample_period * samples_count
            self.current_rt_sample_array = RtSampleArray(
                self._model, observation_time, self._generator.sample_period, samples, self._activation_state)
        return self.current_rt_sample_array

    def set_waveform_generator(self, generator: WaveformGeneratorProtocol):
        self._generator = generator

    @property
    def is_active(self) -> bool:
        return self._activation_state == self._model.pm_types.ComponentActivation.ON


class GenericWaveformProvider:
    """Provide waveform data.

    - runs periodic job to send waveform data
    """

    DEFAULT_WORKER_THREAD_INTERVAL = 0.1  # seconds

    WARN_LIMIT_REALTIMESAMPLES_BEHIND_SCHEDULE = 0.2  # warn limit when real time samples cannot be sent in time
    WARN_RATE_REALTIMESAMPLES_BEHIND_SCHEDULE = 5  # max. every x seconds a message

    def __init__(self, mdib: ProviderMdibProtocol, log_prefix: str = ''):
        self._mdib = mdib
        self._logger = loghelper.get_logger_adapter(f'sdc.device.{self.__class__.__name__}', log_prefix)

        self._stop_worker = Event()
        self._worker_thread: Thread | None = None
        self.notifications_interval = self.DEFAULT_WORKER_THREAD_INTERVAL
        self._waveform_generators: dict[str, _SampleArrayGenerator] = {}
        self._annotators: dict[str, AnnotatorProtocol] = {}
        self._last_log_time = 0
        self._last_logged_delay = 0

    def register_waveform_generator(self, descriptor_handle: str, wf_generator: WaveformGeneratorProtocol):
        """Add wf_generator to waveform sources.

        :param descriptor_handle: the handle of the RealtimeSampleArray that shall accept this data
        :param wf_generator: a waveforms.WaveformGenerator instance
        """
        sample_period = wf_generator.sample_period
        entity = self._mdib.entities.by_handle(descriptor_handle)
        if entity.descriptor.SamplePeriod != sample_period:
            # we must inform subscribers
            entity.descriptor.SamplePeriod = sample_period
            with self._mdib.descriptor_transaction() as mgr:
                mgr.write_entity(entity)

        if descriptor_handle in self._waveform_generators:
            self._waveform_generators[descriptor_handle].set_waveform_generator(wf_generator)
        else:
            self._waveform_generators[descriptor_handle] = _SampleArrayGenerator(self._mdib.data_model,
                                                                                 descriptor_handle,
                                                                                 wf_generator)

    def add_annotation_generator(self,
                                 coded_value: Any,
                                 trigger_handle: str,
                                 annotated_handles: list[str]) -> AnnotatorProtocol:
        """Add annotator to list of annotators."""
        annotation = self._mdib.data_model.pm_types.AnnotationType(coded_value)
        annotator = Annotator(annotation, trigger_handle, annotated_handles)
        self._annotators[annotator.trigger_handle] = annotator
        return annotator

    def start(self):
        """Start worker thread."""
        self._worker_thread = Thread(target=self._worker_thread_loop)
        self._worker_thread.daemon = True
        self._worker_thread.start()

    def stop(self):
        """Stop worker thread."""
        if self._worker_thread is not None:
            self._stop_worker.set()
            self._worker_thread.join()
            self._worker_thread = None

    @property
    def is_running(self) -> bool:
        """Return True if worker thread is set."""
        return self._worker_thread is not None

    def set_activation_state(self, descriptor_handle: str, component_activation_state: ComponentActivation):
        """Set the activation state of waveform generator and of Metric state in mdib."""
        wf_generator = self._waveform_generators[descriptor_handle]
        wf_generator.set_activation_state(component_activation_state)
        with self._mdib.rt_sample_state_transaction() as mgr:
            state = mgr.get_state(descriptor_handle)
            state.ActivationState = component_activation_state
            # if the generator is not active, there shall be no MetricValue
            if not wf_generator.is_active:
                state.MetricValue = None

    def update_all_realtime_samples(self) -> list[EntityProtocol]:
        """Update all realtime sample states that have a waveform generator registered.

        On transaction commit the mdib will call the appropriate send method of the sdc device.
        """
        updated_entities = []
        for descriptor_handle, wf_generator in self._waveform_generators.items():
            if wf_generator.is_active:
                entity = self._mdib.entities.by_handle(descriptor_handle)
                self._update_rt_samples(entity.state)
                updated_entities.append(entity)
        self._add_all_annotations()
        return updated_entities

    def provide_waveforms(self,
                          generator_class: type[WaveformGeneratorProtocol] = waveforms.TriangleGenerator,
                          max_waveforms: int | None = None) -> list[str]:
        """Create waveform generators for waveforms in mdib.

        The values of descriptor.TechnicalRange and descriptor.SamplePeriod are used for the generator.

        :param generator_class: the generator to be instantiated.
        :param max_waveforms: limit number of waveforms.
        :return: list of handles of created generators
        """
        pm_name = self._mdib.data_model.pm_names.RealTimeSampleArrayMetricDescriptor
        all_waveform_entities = self._mdib.entities.by_node_type(pm_name)
        if max_waveforms:
            all_waveform_entities = all_waveform_entities[:max_waveforms]
        for entity in all_waveform_entities:
            min_value = 0
            max_value = 1
            # guarantee usable value:
            sample_period = entity.descriptor.SamplePeriod if entity.descriptor.SamplePeriod > 0 else 0.01
            try:
                tech_range = entity.descriptor.TechnicalRange[0]
            except IndexError:
                pass
            else:
                if tech_range.Lower is not None:
                    min_value = float(tech_range.Lower)
                if tech_range.Upper is not None:
                    max_value = float(tech_range.Upper)
                if max_value < min_value:
                    max_value, min_value = min_value, max_value  # swap values
                elif min_value == max_value:
                    max_value += 1
            generator = generator_class(min_value=min_value,
                                        max_value=max_value,
                                        waveform_period=2.0,
                                        sample_period=sample_period)
            self.register_waveform_generator(entity.handle, generator)
        return [ent.handle for ent in all_waveform_entities]

    def _worker_thread_loop(self):
        timer = IntervalTimer(period_in_seconds=self.notifications_interval)
        try:
            while True:
                shall_stop = self._stop_worker.is_set()
                if shall_stop:
                    return
                behind_schedule_seconds = timer.wait_next_interval_begin()
                self._log_waveform_timing(behind_schedule_seconds)
                try:
                    updated_entities =  self.update_all_realtime_samples()
                    with self._mdib.rt_sample_state_transaction() as transaction:
                        transaction.write_entities(updated_entities)
                    self._log_waveform_timing(behind_schedule_seconds)
                except Exception:  # noqa: BLE001
                    # catch all to keep loop running
                    self._logger.warning('could not update real time samples: %s', traceback.format_exc())
        finally:
            self._logger.info('rt_sample_sendloop end')

    def _update_rt_samples(self, state: RealTimeSampleArrayMetricStateContainer):
        """Update waveforms state from waveform generator (if available)."""
        wf_generator = self._waveform_generators.get(state.DescriptorHandle)
        if wf_generator:
            ctxt = Context(prec=10)
            rt_sample_array = wf_generator.get_next_sample_array()
            samples = [ctxt.create_decimal(s) for s in rt_sample_array.samples]
            if state.MetricValue is None:
                state.mk_metric_value()
            state.MetricValue.Samples = samples
            state.MetricValue.DeterminationTime = rt_sample_array.determination_time
            state.MetricValue.Annotation = rt_sample_array.annotations
            state.MetricValue.ApplyAnnotation = rt_sample_array.apply_annotations
            state.ActivationState = rt_sample_array.activation_state

    def _add_all_annotations(self):
        """Add annotations to all current RtSampleArrays."""
        rt_sample_arrays = {handle: g.current_rt_sample_array for (handle, g) in self._waveform_generators.items()}
        for src_handle, _annotator in self._annotators.items():
            if src_handle in rt_sample_arrays:
                timestamps = _annotator.get_annotation_timestamps(rt_sample_arrays[src_handle])
                if timestamps:
                    for dest_handle in _annotator.annotated_handles:
                        if dest_handle in rt_sample_arrays:
                            rt_sample_arrays[dest_handle].add_annotations_at(_annotator.annotation, timestamps)

    def _log_waveform_timing(self, behind_schedule_seconds: float):
        try:
            last_log_time = self._last_log_time
        except AttributeError:
            self._last_log_time = 0
            last_log_time = self._last_log_time
        try:
            last_logged_delay = self._last_logged_delay
        except AttributeError:
            self._last_logged_delay = 0
            last_logged_delay = self._last_logged_delay

        # max. one log per second
        now = time.monotonic()
        if now - last_log_time < self.WARN_RATE_REALTIMESAMPLES_BEHIND_SCHEDULE:
            return
        if last_logged_delay >= self.WARN_LIMIT_REALTIMESAMPLES_BEHIND_SCHEDULE > behind_schedule_seconds:
            self._logger.info('RealTimeSampleTimer delay is back inside limit of %2f seconds (mdib version=%d)',
                              self.WARN_LIMIT_REALTIMESAMPLES_BEHIND_SCHEDULE, self._mdib.mdib_version)
            self._last_logged_delay = behind_schedule_seconds
            self._last_log_time = now
        elif behind_schedule_seconds >= self.WARN_LIMIT_REALTIMESAMPLES_BEHIND_SCHEDULE:
            self._logger.warning('RealTimeSampleTimer is %.4f seconds behind schedule (mdib version=%d)',
                                 behind_schedule_seconds, self._mdib.mdib_version)
            self._last_logged_delay = behind_schedule_seconds
            self._last_log_time = now



"""Definitions of descriptor containers."""
from __future__ import annotations

import inspect
import sys
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, ClassVar, Protocol

from sdc11073 import observableproperties as properties
from sdc11073.xml_types import ext_qnames as ext
from sdc11073.xml_types import msg_qnames as msg
from sdc11073.xml_types import pm_qnames, pm_types
from sdc11073.xml_types import xml_structure as x_struct

from .containerbase import ContainerBase

if TYPE_CHECKING:
    from collections.abc import Iterable
    from decimal import Decimal

    from lxml import etree

    from sdc11073 import xml_utils
    from sdc11073.xml_types.isoduration import DurationType
    from sdc11073.xml_types.xml_structure import ExtensionLocalValue


@dataclass(frozen=True)
class ChildDescriptorMapping:
    """Map element names to node types.

    Needed when building a xml tree of descriptors.
    The name of a child element is often not identical to the type of the descriptor, e.g. a channel uses
    pm.Metric for all classes derived from AbstractMetricDescriptor.
    """

    child_qname: etree.QName
    node_types: tuple[etree.QName, ...] = None

    def __repr__(self) -> str:
        if self.node_types is None:
            return f'{self.__class__.__name__} name={self.child_qname.localname} (no types)'
        types = ', '.join([t.localname for t in self.node_types])
        return f'{self.__class__.__name__} name={self.child_qname.localname} types={types}'


def sorted_child_data(obj: Any, member_name: str):
    """:return: an iterator with whatever the members have, starting with base class members"""
    classes = inspect.getmro(obj.__class__)
    for cls in reversed(classes):
        names = cls.__dict__.get(member_name, [])
        yield from names


class AbstractDescriptorProtocol(Protocol):
    """The common Interface of all descriptors."""

    NODETYPE: etree.QName
    STATE_QNAME: etree.QName
    is_descriptor_container: bool
    is_system_context_descriptor: bool
    is_realtime_sample_array_metric_descriptor: bool
    is_metric_descriptor: bool
    is_operational_descriptor: bool
    is_component_descriptor: bool
    is_alert_descriptor: bool
    is_alert_signal_descriptor: bool
    is_alert_condition_descriptor: bool
    is_context_descriptor: bool
    is_leaf: bool  # determines if children can be added
    Handle: str
    DescriptorVersion: int
    SafetyClassification: pm_types.SafetyClassification
    Type: pm_types.CodedValue | None
    source_mds: str
    parent_handle: str | None
    coding: pm_types.Coding | None

    def __init__(self, handle: str, parent_handle: str | None):
        ...

    def set_source_mds(self, handle: str):
        """Set source_mds member."""
        ...

    @classmethod
    def from_node(cls, node: xml_utils.LxmlElement, parent_handle: str | None = None) -> AbstractDescriptorProtocol:
        """Create class and init its properties from the node."""


class AbstractDescriptorContainer(ContainerBase):
    """AbstractDescriptorContainer represents the AbstractDescriptor of BICEPS."""

    # these class variables allow easy type-checking. Derived classes will set corresponding values to True
    # pylint: disable=invalid-name
    is_descriptor_container = True
    is_system_context_descriptor = False
    is_realtime_sample_array_metric_descriptor = False
    is_metric_descriptor = False
    is_operational_descriptor = False
    is_component_descriptor = False
    is_alert_descriptor = False
    is_alert_signal_descriptor = False
    is_alert_condition_descriptor = False
    is_context_descriptor = False

    is_leaf = True  # determines if children can be added
    node = properties.ObservableProperty()  # the etree node

    Handle: str = x_struct.HandleAttributeProperty('Handle', is_optional=False)
    Extension: ExtensionLocalValue = x_struct.ExtensionNodeProperty(ext.Extension)
    DescriptorVersion: int = x_struct.VersionCounterAttributeProperty('DescriptorVersion',
                                                                      implied_py_value=0)
    SafetyClassification: pm_types.SafetyClassification = x_struct.EnumAttributeProperty('SafetyClassification',
                                                                                         implied_py_value=pm_types.SafetyClassification.INF,
                                                                                         enum_cls=pm_types.SafetyClassification)
    Type: pm_types.CodedValue | None = x_struct.SubElementProperty(pm_qnames.Type,
                                                                   value_class=pm_types.CodedValue,
                                                                   is_optional=True)
    # pylint: enable=invalid-name
    _props = ('Handle', 'DescriptorVersion', 'SafetyClassification', 'Extension', 'Type')
    _child_elements_order = (ext.Extension, pm_qnames.Type)  # child elements in BICEPS order
    STATE_QNAME = None
    extension_class_lookup: ClassVar[dict[etree.QName, type[pm_types.PropertyBasedPMType]]] = {
        msg.Retrievability: pm_types.Retrievability,
    }

    def __init__(self, handle: str | None, parent_handle: str | None):
        """Parent Handle can only be None for a Mds Descriptor. every other descriptor has a parent."""
        super().__init__()
        self._parent_handle = parent_handle
        self.Handle = handle
        self._source_mds = None  # needed on device side if mdib contains > 1 mds

    @property
    def coding(self) -> pm_types.Coding | None:  # noqa: D102
        return self.Type.coding if self.Type is not None else None

    @property
    def code_id(self) -> str | None:  # noqa: D102
        return self.Type.Code if self.Type is not None else None  # pylint:disable=no-member

    @property
    def coding_system(self) -> str | None:  # noqa: D102
        return self.Type.CodingSystem if self.Type is not None else None  # pylint:disable=no-member

    @property
    def parent_handle(self) -> str | None:  # noqa: D102
        return self._parent_handle

    @parent_handle.setter
    def parent_handle(self, value: str):
        self._parent_handle = value

    def get_retrievability(self) -> list[pm_types.Retrievability]:
        """Return all retrievability data from Extension."""
        return [pm_types.Retrievability.from_node(x) for x in self.Extension if x.tag == msg.Retrievability]

    def set_retrievability(self, retrievabilities: Iterable[pm_types.Retrievability]) -> None:
        """Replace all retrievability elements in Extension with provided ones."""
        for tmp in [x for x in self.Extension if x.tag == msg.Retrievability]:
            self.Extension.remove(tmp)
        self.Extension.extend([r.as_etree_node(msg.Retrievability, {}) for r in retrievabilities])

    def increment_descriptor_version(self):
        """Increment DescriptorVersion."""
        self.DescriptorVersion += 1

    def update_from_other_container(self, other: AbstractDescriptorContainer,
                                    skipped_properties: list[str] | None = None):
        """Update own properties with values from other descriptor."""
        if other.Handle != self.Handle:
            msg = (f'Update from a container with different handle is not possible! '
                  f'Have "{self.Handle}", got "{other.Handle}"')
            raise ValueError(msg)
        self._update_from_other(other, skipped_properties)
        self.node = other.node

    def get_actual_value(self, attr_name: str) -> Any:
        """Ignores default value and implied value, e.g. returns None if value is not present in xml."""
        return getattr(self.__class__, attr_name).get_actual_value(self)

    def diff(self, other: AbstractDescriptorContainer, ignore_property_names: list[str] | None = None) -> None | list[
        str]:
        """Compare with another descriptor.

        It compares all properties plus the parent handle member.
        :param other: the object (descriptor container) to compare with
        :param ignore_property_names: list of properties that shall be excluded from diff calculation
        :return: textual representation of differences or None if equal.
        """
        ret = super().diff(other, ignore_property_names) or []
        if ignore_property_names is None or 'parent_handle' not in ignore_property_names:
            my_value = self.parent_handle
            try:
                other_value = other.parent_handle
            except AttributeError:
                ret.append(f'parent_handle={my_value}, other does not have this attribute')
            else:
                if my_value != other_value:
                    ret.append(f'parent_handle={my_value}, other={other_value}')
        return None if len(ret) == 0 else ret

    def tag_name_for_child_descriptor(self, node_type: etree.QName) -> (etree.QName, bool):
        """Determine the tag name of a child descriptor.

        This isneeded when the xml tree of the descriptor is created.
        It uses the _child_elements_order members of the class itself and its base classes
        which map node type to tag name.
        :param node_type: the type QName (NODETYPE member)
        :return: tuple(QName, set_xsi_type_flag).
        """
        for child in sorted_child_data(self, '_child_descriptor_name_mappings'):
            if child.node_types is not None and node_type in child.node_types:
                set_xsi_type = len(child.node_types) > 1
                return child.child_qname, set_xsi_type
        msg = f'{node_type} not known in child declarations of {self.__class__.__name__}'
        raise ValueError(msg)

    def sort_child_nodes(self, node: xml_utils.LxmlElement) -> None:
        """Bring all child elements of node in correct order (BICEPS schema).

        raises a ValueError if a child node exist that is not listed in ordered_tags
        :param node: the element to be sorted.
        """
        q_names = list(sorted_child_data(self, '_child_elements_order'))
        not_in_order = [n for n in node if n.tag not in q_names]
        if len(not_in_order) > 0:
            msg = (f'{self.__class__.__name__}: not in Order:{[n.tag for n in not_in_order]} '
                  f'node={node.tag}, order={[o.localname for o in q_names]}')
            raise ValueError(msg)
        all_child_nodes = node[:]
        for child_node in all_child_nodes:
            node.remove(child_node)
        for q_name in q_names:
            for child_node in all_child_nodes:
                if child_node.tag == q_name:
                    node.append(child_node)

    def set_source_mds(self, handle: str):
        """Set source_mds member."""
        self._source_mds = handle

    @property
    def source_mds(self) -> str:
        """Get source mds handle."""
        return self._source_mds

    def __str__(self) -> str:
        name = self.NODETYPE.localname or None
        return (f'Descriptor "{name}": handle={self.Handle} descriptor version={self.DescriptorVersion} '
                f'parent handle={self.parent_handle}')

    def __repr__(self) -> str:
        name = self.NODETYPE.localname or None
        return (f'Descriptor "{name}": handle={self.Handle} descriptor version={self.DescriptorVersion} '
                f'parent={self.parent_handle}')

    @classmethod
    def from_node(cls, node: xml_utils.LxmlElement, parent_handle: str | None = None) -> AbstractDescriptorContainer:
        """Create class and init its properties from the node."""
        obj = cls(handle=None,  # will be determined in constructor from node value
                  parent_handle=parent_handle)
        obj.update_from_node(node)
        return obj


class AbstractDeviceComponentDescriptorContainer(AbstractDescriptorContainer):
    """Represents AbstractDeviceComponentDescriptor in BICEPS."""

    is_component_descriptor = True
    is_leaf = False
    ProductionSpecification: list[pm_types.ProductionSpecification] = x_struct.SubElementListProperty(
        pm_qnames.ProductionSpecification,
        value_class=pm_types.ProductionSpecification)
    _props = ('ProductionSpecification',)
    _child_elements_order = (pm_qnames.ProductionSpecification,)


class AbstractComplexDeviceComponentDescriptorContainer(AbstractDeviceComponentDescriptorContainer):
    """Represents AbstractComplexDeviceComponentDescriptor in BICEPS."""

    _child_elements_order = (pm_qnames.AlertSystem, pm_qnames.Sco)
    _child_descriptor_name_mappings = (
        ChildDescriptorMapping(pm_qnames.AlertSystem, (pm_qnames.AlertSystemDescriptor,)),
        ChildDescriptorMapping(pm_qnames.Sco, (pm_qnames.ScoDescriptor,)))


class MdsDescriptorContainer(AbstractComplexDeviceComponentDescriptorContainer):
    """Represents MdsDescriptor in BICEPS."""

    NODETYPE = pm_qnames.MdsDescriptor
    STATE_QNAME = pm_qnames.MdsState
    MetaData = x_struct.SubElementProperty(pm_qnames.MetaData, value_class=pm_types.MetaData, is_optional=True)
    ApprovedJurisdictions: pm_types.ApprovedJurisdictions | None = x_struct.SubElementProperty(
        pm_qnames.ApprovedJurisdictions,
        value_class=pm_types.ApprovedJurisdictions,
        is_optional=True)
    _props = ('MetaData', 'ApprovedJurisdictions')
    _child_elements_order = (pm_qnames.MetaData,
                             pm_qnames.SystemContext,
                             pm_qnames.Clock,
                             pm_qnames.Battery,
                             pm_qnames.ApprovedJurisdictions,
                             pm_qnames.Vmd)
    _child_descriptor_name_mappings = (
        ChildDescriptorMapping(pm_qnames.SystemContext, (pm_qnames.SystemContextDescriptor,)),
        ChildDescriptorMapping(pm_qnames.Clock, (pm_qnames.ClockDescriptor,)),
        ChildDescriptorMapping(pm_qnames.Battery, (pm_qnames.BatteryDescriptor,)),
        ChildDescriptorMapping(pm_qnames.Vmd, (pm_qnames.VmdDescriptor,)),
    )


class VmdDescriptorContainer(AbstractComplexDeviceComponentDescriptorContainer):
    """Represents VmdDescriptor in BICEPS."""

    NODETYPE = pm_qnames.VmdDescriptor
    STATE_QNAME = pm_qnames.VmdState

    ApprovedJurisdictions: pm_types.ApprovedJurisdictions | None = x_struct.SubElementProperty(
        pm_qnames.ApprovedJurisdictions,
        value_class=pm_types.ApprovedJurisdictions,
        is_optional=True)
    _props = ('ApprovedJurisdictions',)
    _child_elements_order = (pm_qnames.ApprovedJurisdictions,
                             pm_qnames.Channel)
    _child_descriptor_name_mappings = (
        ChildDescriptorMapping(pm_qnames.Channel, (pm_qnames.ChannelDescriptor,)),
    )


class ChannelDescriptorContainer(AbstractDeviceComponentDescriptorContainer):
    """Represents ChannelDescriptor in BICEPS."""

    NODETYPE = pm_qnames.ChannelDescriptor
    STATE_QNAME = pm_qnames.ChannelState
    _child_elements_order = (pm_qnames.Metric,)
    _child_descriptor_name_mappings = (
        ChildDescriptorMapping(pm_qnames.Metric, (pm_qnames.NumericMetricDescriptor,
                                                  pm_qnames.StringMetricDescriptor,
                                                  pm_qnames.EnumStringMetricDescriptor,
                                                  pm_qnames.RealTimeSampleArrayMetricDescriptor,
                                                  pm_qnames.DistributionSampleArrayMetricDescriptor,
                                                  ),
                               ),
    )


class ClockDescriptorContainer(AbstractDeviceComponentDescriptorContainer):
    """Represents ClockDescriptor in BICEPS."""

    NODETYPE = pm_qnames.ClockDescriptor
    STATE_QNAME = pm_qnames.ClockState
    TimeProtocol: list[pm_types.CodedValue] = x_struct.SubElementListProperty(pm_qnames.TimeProtocol,
                                                                              value_class=pm_types.CodedValue)
    Resolution: DurationType | None = x_struct.DurationAttributeProperty('Resolution')
    _props = ('TimeProtocol', 'Resolution')
    _child_elements_order = (pm_qnames.TimeProtocol,)


class BatteryDescriptorContainer(AbstractDeviceComponentDescriptorContainer):
    """Represents BatteryDescriptor in BICEPS."""

    NODETYPE = pm_qnames.BatteryDescriptor
    STATE_QNAME = pm_qnames.BatteryState
    # pylint: disable=invalid-name
    CapacityFullCharge: pm_types.Measurement | None = x_struct.SubElementProperty(pm_qnames.CapacityFullCharge,
                                                                                  value_class=pm_types.Measurement,
                                                                                  is_optional=True)
    CapacitySpecified: pm_types.Measurement | None = x_struct.SubElementProperty(pm_qnames.CapacitySpecified,
                                                                                 value_class=pm_types.Measurement,
                                                                                 is_optional=True)
    VoltageSpecified: pm_types.Measurement | None = x_struct.SubElementProperty(pm_qnames.VoltageSpecified,
                                                                                value_class=pm_types.Measurement,
                                                                                is_optional=True)
    # pylint: enable=invalid-name
    _props = ('CapacityFullCharge', 'CapacitySpecified', 'VoltageSpecified')
    _child_elements_order = (pm_qnames.CapacityFullCharge,
                             pm_qnames.CapacitySpecified,
                             pm_qnames.VoltageSpecified)


class ScoDescriptorContainer(AbstractDeviceComponentDescriptorContainer):
    """Represents ScoDescriptor in BICEPS."""

    NODETYPE = pm_qnames.ScoDescriptor
    STATE_QNAME = pm_qnames.ScoState
    _child_elements_order = (pm_qnames.Operation,)
    _child_descriptor_name_mappings = (
        ChildDescriptorMapping(pm_qnames.Operation, (pm_qnames.SetValueOperationDescriptor,
                                                     pm_qnames.SetStringOperationDescriptor,
                                                     pm_qnames.SetContextStateOperationDescriptor,
                                                     pm_qnames.SetMetricStateOperationDescriptor,
                                                     pm_qnames.SetComponentStateOperationDescriptor,
                                                     pm_qnames.SetAlertStateOperationDescriptor,
                                                     pm_qnames.ActivateOperationDescriptor,
                                                     ),
                               ),
    )


class AbstractMetricDescriptorContainer(AbstractDescriptorContainer):
    """Represents AbstractMetricDescriptor in BICEPS."""

    is_metric_descriptor = True
    Unit: pm_types.CodedValue = x_struct.SubElementProperty(pm_qnames.Unit, value_class=pm_types.CodedValue)
    BodySite: list[pm_types.CodedValue] = x_struct.SubElementListProperty(pm_qnames.BodySite,
                                                                          value_class=pm_types.CodedValue)
    Relation: list[pm_types.Relation] = x_struct.SubElementListProperty(pm_qnames.Relation,
                                                                        value_class=pm_types.Relation)
    MetricCategory: pm_types.MetricCategory = x_struct.EnumAttributeProperty(
        'MetricCategory',
        enum_cls=pm_types.MetricCategory,
        default_py_value=pm_types.MetricCategory.UNSPECIFIED,
        is_optional=False)
    DerivationMethod: pm_types.DerivationMethod | None = x_struct.EnumAttributeProperty('DerivationMethod',
                                                                                        enum_cls=pm_types.DerivationMethod)
    #  There is an implied value defined, but it is complicated, therefore here not implemented:
    # - If pm:AbstractDescriptor/@MetricCategory is "Set" or "Preset",
    #         then the default value of DerivationMethod is "Man"
    # - If pm:AbstractDescriptor/@MetricCategory is "Clc", "Msrmt", "Rcmm",
    #         then the default value of DerivationMethod is "Auto"
    # - If pm:AbstractDescriptor/@MetricCategory is "Unspec",
    #         then no default value is being implied</xsd:documentation>
    MetricAvailability: pm_types.MetricAvailability = x_struct.EnumAttributeProperty(
        'MetricAvailability',
        enum_cls=pm_types.MetricAvailability,
        default_py_value=pm_types.MetricAvailability.CONTINUOUS,
        is_optional=False)
    MaxMeasurementTime: DurationType | None = x_struct.DurationAttributeProperty('MaxMeasurementTime')
    MaxDelayTime: DurationType | None = x_struct.DurationAttributeProperty('MaxDelayTime')
    DeterminationPeriod: DurationType | None = x_struct.DurationAttributeProperty('DeterminationPeriod')
    LifeTimePeriod: DurationType | None = x_struct.DurationAttributeProperty('LifeTimePeriod')
    ActivationDuration: DurationType | None = x_struct.DurationAttributeProperty('ActivationDuration')
    _props = (
        'Unit', 'BodySite', 'Relation', 'MetricCategory', 'DerivationMethod', 'MetricAvailability',
        'MaxMeasurementTime',
        'MaxDelayTime', 'DeterminationPeriod', 'LifeTimePeriod', 'ActivationDuration')
    _child_elements_order = (pm_qnames.Unit,
                             pm_qnames.BodySite,
                             pm_qnames.Relation)


class NumericMetricDescriptorContainer(AbstractMetricDescriptorContainer):
    """Represents NumericMetricDescriptor in BICEPS."""

    NODETYPE = pm_qnames.NumericMetricDescriptor
    STATE_QNAME = pm_qnames.NumericMetricState
    TechnicalRange: list[pm_types.Range] = x_struct.SubElementListProperty(pm_qnames.TechnicalRange,
                                                                           value_class=pm_types.Range)
    Resolution: Decimal = x_struct.DecimalAttributeProperty('Resolution', is_optional=False)
    AveragingPeriod: DurationType | None = x_struct.DurationAttributeProperty('AveragingPeriod')
    _props = ('TechnicalRange', 'Resolution', 'AveragingPeriod')
    _child_elements_order = (pm_qnames.TechnicalRange,)


class StringMetricDescriptorContainer(AbstractMetricDescriptorContainer):
    """Represents StringMetricDescriptor in BICEPS."""

    NODETYPE = pm_qnames.StringMetricDescriptor
    STATE_QNAME = pm_qnames.StringMetricState


class EnumStringMetricDescriptorContainer(StringMetricDescriptorContainer):
    """Represents EnumStringMetricDescriptor in BICEPS."""

    NODETYPE = pm_qnames.EnumStringMetricDescriptor
    STATE_QNAME = pm_qnames.EnumStringMetricState
    AllowedValue: list[pm_types.AllowedValue] = x_struct.SubElementListProperty(pm_qnames.AllowedValue,
                                                                                value_class=pm_types.AllowedValue)
    _props = ('AllowedValue',)
    _child_elements_order = (pm_qnames.AllowedValue,)


class RealTimeSampleArrayMetricDescriptorContainer(AbstractMetricDescriptorContainer):
    """Represents RealTimeSampleArrayMetricDescriptor in BICEPS."""

    is_realtime_sample_array_metric_descriptor = True
    NODETYPE = pm_qnames.RealTimeSampleArrayMetricDescriptor
    STATE_QNAME = pm_qnames.RealTimeSampleArrayMetricState
    TechnicalRange: list[pm_types.Range] = x_struct.SubElementListProperty(pm_qnames.TechnicalRange,
                                                                           value_class=pm_types.Range)
    Resolution: Decimal = x_struct.DecimalAttributeProperty('Resolution', is_optional=False)
    SamplePeriod: DurationType = x_struct.DurationAttributeProperty('SamplePeriod', is_optional=False)
    _props = ('TechnicalRange', 'Resolution', 'SamplePeriod')
    _child_elements_order = (pm_qnames.TechnicalRange,)


class DistributionSampleArrayMetricDescriptorContainer(AbstractMetricDescriptorContainer):
    """Represents DistributionSampleArrayMetricDescriptor in BICEPS."""

    NODETYPE = pm_qnames.DistributionSampleArrayMetricDescriptor
    STATE_QNAME = pm_qnames.DistributionSampleArrayMetricState
    TechnicalRange: list[pm_types.Range] = x_struct.SubElementListProperty(pm_qnames.TechnicalRange,
                                                                           value_class=pm_types.Range)
    DomainUnit: pm_types.CodedValue = x_struct.SubElementProperty(pm_qnames.DomainUnit, value_class=pm_types.CodedValue)
    DistributionRange: pm_types.Range = x_struct.SubElementProperty(pm_qnames.DistributionRange,
                                                                    value_class=pm_types.Range,
                                                                    default_py_value=pm_types.Range())
    Resolution: Decimal = x_struct.DecimalAttributeProperty('Resolution', is_optional=False)
    _props = ('TechnicalRange', 'DomainUnit', 'DistributionRange', 'Resolution')
    _child_elements_order = (pm_qnames.TechnicalRange,
                             pm_qnames.DomainUnit,
                             pm_qnames.DistributionRange)


class AbstractOperationDescriptorContainer(AbstractDescriptorContainer):
    """Represents AbstractOperationDescriptor in BICEPS."""

    is_operational_descriptor = True
    OperationTarget: str = x_struct.HandleRefAttributeProperty('OperationTarget', is_optional=False)
    MaxTimeToFinish: DurationType | None = x_struct.DurationAttributeProperty('MaxTimeToFinish')
    InvocationEffectiveTimeout: DurationType | None = x_struct.DurationAttributeProperty(
        'InvocationEffectiveTimeout')
    Retriggerable: bool = x_struct.BooleanAttributeProperty('Retriggerable', implied_py_value=True)
    AccessLevel: pm_types.AccessLevel = x_struct.EnumAttributeProperty('AccessLevel',
                                                                       implied_py_value=pm_types.AccessLevel.USER,
                                                                       enum_cls=pm_types.AccessLevel)
    _props = ('OperationTarget', 'MaxTimeToFinish', 'InvocationEffectiveTimeout', 'Retriggerable', 'AccessLevel')


class AbstractOperationDescriptorProtocol(AbstractDescriptorProtocol):
    """Protocol definition for AbstractOperationDescriptorContainer."""

    OperationTarget: str
    MaxTimeToFinish: DurationType | None
    InvocationEffectiveTimeout: DurationType | None
    Retriggerable: bool
    AccessLevel: pm_types.AccessLevel


class SetValueOperationDescriptorContainer(AbstractOperationDescriptorContainer):
    """Represents SetValueOperationDescriptor in BICEPS."""

    NODETYPE = pm_qnames.SetValueOperationDescriptor
    STATE_QNAME = pm_qnames.SetValueOperationState


class SetStringOperationDescriptorContainer(AbstractOperationDescriptorContainer):
    """Represents SetStringOperationDescriptor in BICEPS."""

    NODETYPE = pm_qnames.SetStringOperationDescriptor
    STATE_QNAME = pm_qnames.SetStringOperationState
    MaxLength: int | None = x_struct.IntegerAttributeProperty('MaxLength')
    _props = ('MaxLength',)


class AbstractSetStateOperationDescriptorContainer(AbstractOperationDescriptorContainer):
    """Represents AbstractSetStateOperationDescriptor in BICEPS."""

    ModifiableData: list[str] = x_struct.SubElementStringListProperty(pm_qnames.ModifiableData)
    _props = ('ModifiableData',)
    _child_elements_order = (pm_qnames.ModifiableData,)


class SetContextStateOperationDescriptorContainer(AbstractSetStateOperationDescriptorContainer):
    """Represents SetContextStateOperationDescriptor in BICEPS."""

    NODETYPE = pm_qnames.SetContextStateOperationDescriptor
    STATE_QNAME = pm_qnames.SetContextStateOperationState


class SetMetricStateOperationDescriptorContainer(AbstractSetStateOperationDescriptorContainer):
    """Represents SetMetricStateOperationDescriptor in BICEPS."""

    NODETYPE = pm_qnames.SetMetricStateOperationDescriptor
    STATE_QNAME = pm_qnames.SetMetricStateOperationState


class SetComponentStateOperationDescriptorContainer(AbstractSetStateOperationDescriptorContainer):
    """Represents SetComponentStateOperationDescriptor in BICEPS."""

    NODETYPE = pm_qnames.SetComponentStateOperationDescriptor
    STATE_QNAME = pm_qnames.SetComponentStateOperationState


class SetAlertStateOperationDescriptorContainer(AbstractSetStateOperationDescriptorContainer):
    """Represents SetAlertStateOperationDescriptor in BICEPS."""

    NODETYPE = pm_qnames.SetAlertStateOperationDescriptor
    STATE_QNAME = pm_qnames.SetAlertStateOperationState


class ActivateOperationDescriptorContainer(AbstractSetStateOperationDescriptorContainer):
    """Represents ActivateOperationDescriptor in BICEPS."""

    NODETYPE = pm_qnames.ActivateOperationDescriptor
    STATE_QNAME = pm_qnames.ActivateOperationState
    Argument: list[pm_types.ActivateOperationDescriptorArgument] = x_struct.SubElementListProperty(
        pm_qnames.Argument, value_class=pm_types.ActivateOperationDescriptorArgument)
    _props = ('Argument',)
    _child_elements_order = (pm_qnames.Argument,)


class AbstractAlertDescriptorContainer(AbstractDescriptorContainer):
    """Represents AbstractAlertDescriptor in BICEPS."""

    is_alert_descriptor = True
    is_leaf = False


class AlertSystemDescriptorContainer(AbstractAlertDescriptorContainer):
    """Represents AlertSystemDescriptor in BICEPS.

    AlertSystemDescriptor describes an Alert system to detect Alert conditions and generate Alert signals,
    which belong to specific Alert conditions.
    Alert conditions are represented by a list of pm:AlertConditionDescriptor elements and
    Alert signals are represented by a list of pm:AlertSignalDescriptor elements.
    """

    NODETYPE = pm_qnames.AlertSystemDescriptor
    STATE_QNAME = pm_qnames.AlertSystemState
    MaxPhysiologicalParallelAlarms: int | None = x_struct.UnsignedIntAttributeProperty('MaxPhysiologicalParallelAlarms')
    MaxTechnicalParallelAlarms: int | None = x_struct.UnsignedIntAttributeProperty('MaxTechnicalParallelAlarms')
    SelfCheckPeriod: DurationType | None = x_struct.DurationAttributeProperty('SelfCheckPeriod')
    _props = ('MaxPhysiologicalParallelAlarms', 'MaxTechnicalParallelAlarms', 'SelfCheckPeriod')
    _child_elements_order = (pm_qnames.AlertCondition,
                             pm_qnames.AlertSignal)
    _child_descriptor_name_mappings = (
        ChildDescriptorMapping(pm_qnames.AlertCondition, (pm_qnames.AlertConditionDescriptor,
                                                          pm_qnames.LimitAlertConditionDescriptor,
                                                          ),
                               ),
        ChildDescriptorMapping(pm_qnames.AlertSignal, (pm_qnames.AlertSignalDescriptor,)),
    )


class AlertConditionDescriptorContainer(AbstractAlertDescriptorContainer):
    """Represents AlertConditionDescriptor in BICEPS.

    An ALERT CONDITION contains the information about a potentially or actually HAZARDOUS SITUATION.
    Examples: a physiological alarm limit has been exceeded or a sensor has been unplugged.
    """

    is_alert_condition_descriptor = True
    NODETYPE = pm_qnames.AlertConditionDescriptor
    STATE_QNAME = pm_qnames.AlertConditionState
    Source: list[str] = x_struct.SubElementHandleRefListProperty(
        pm_qnames.Source)  # a list of 0...n pm:HandleRef elements
    CauseInfo: list[pm_types.CauseInfo] = x_struct.SubElementListProperty(pm_qnames.CauseInfo,
                                                                          value_class=pm_types.CauseInfo)
    Kind: pm_types.AlertConditionKind = x_struct.EnumAttributeProperty('Kind',
                                                                       default_py_value=pm_types.AlertConditionKind.OTHER,
                                                                       enum_cls=pm_types.AlertConditionKind,
                                                                       is_optional=False)
    Priority: pm_types.AlertConditionPriority = x_struct.EnumAttributeProperty('Priority',
                                                                               default_py_value=pm_types.AlertConditionPriority.NONE,
                                                                               enum_cls=pm_types.AlertConditionPriority,
                                                                               is_optional=False)
    DefaultConditionGenerationDelay: DurationType = x_struct.DurationAttributeProperty(
        'DefaultConditionGenerationDelay',
        implied_py_value=0)
    CanEscalate: pm_types.CanEscalate | None = x_struct.EnumAttributeProperty(
        'CanEscalate', enum_cls=pm_types.CanEscalate)
    CanDeescalate: pm_types.CanDeEscalate | None = x_struct.EnumAttributeProperty(
        'CanDeescalate', enum_cls=pm_types.CanDeEscalate)
    _props = ('Source', 'CauseInfo', 'Kind', 'Priority', 'DefaultConditionGenerationDelay',
              'CanEscalate', 'CanDeescalate')
    _child_elements_order = (pm_qnames.Source,
                             pm_qnames.CauseInfo)


class LimitAlertConditionDescriptorContainer(AlertConditionDescriptorContainer):
    """Represents AlertConditionDescriptor in BICEPS."""

    NODETYPE = pm_qnames.LimitAlertConditionDescriptor
    STATE_QNAME = pm_qnames.LimitAlertConditionState
    MaxLimits: pm_types.Range = x_struct.SubElementProperty(pm_qnames.MaxLimits,
                                                            value_class=pm_types.Range,
                                                            default_py_value=pm_types.Range())
    AutoLimitSupported: bool = x_struct.BooleanAttributeProperty('AutoLimitSupported', implied_py_value=False)
    _props = ('MaxLimits', 'AutoLimitSupported')
    _child_elements_order = (pm_qnames.MaxLimits,)


class AlertSignalDescriptorContainer(AbstractAlertDescriptorContainer):
    """Represents AlertSignalDescriptor in BICEPS."""

    is_alert_signal_descriptor = True
    NODETYPE = pm_qnames.AlertSignalDescriptor
    STATE_QNAME = pm_qnames.AlertSignalState
    ConditionSignaled: str = x_struct.HandleRefAttributeProperty('ConditionSignaled')
    Manifestation: pm_types.AlertSignalManifestation = x_struct.EnumAttributeProperty('Manifestation',
                                                                                      enum_cls=pm_types.AlertSignalManifestation,
                                                                                      is_optional=False)
    Latching: bool = x_struct.BooleanAttributeProperty('Latching', default_py_value=False, is_optional=False)
    DefaultSignalGenerationDelay: DurationType = x_struct.DurationAttributeProperty(
        'DefaultSignalGenerationDelay',
        implied_py_value=0)
    MinSignalGenerationDelay: DurationType | None = x_struct.DurationAttributeProperty(
        'MinSignalGenerationDelay')
    MaxSignalGenerationDelay: DurationType | None = x_struct.DurationAttributeProperty(
        'MaxSignalGenerationDelay')
    SignalDelegationSupported: bool = x_struct.BooleanAttributeProperty('SignalDelegationSupported',
                                                                        implied_py_value=False)
    AcknowledgementSupported: bool = x_struct.BooleanAttributeProperty('AcknowledgementSupported',
                                                                       implied_py_value=False)
    AcknowledgeTimeout: DurationType | None = x_struct.DurationAttributeProperty('AcknowledgeTimeout')
    _props = ('ConditionSignaled', 'Manifestation', 'Latching', 'DefaultSignalGenerationDelay',
              'MinSignalGenerationDelay', 'MaxSignalGenerationDelay',
              'SignalDelegationSupported', 'AcknowledgementSupported', 'AcknowledgeTimeout')


class SystemContextDescriptorContainer(AbstractDeviceComponentDescriptorContainer):
    """Represents SystemContextDescriptor in BICEPS."""

    is_system_context_descriptor = True
    NODETYPE = pm_qnames.SystemContextDescriptor
    STATE_QNAME = pm_qnames.SystemContextState
    _child_elements_order = (pm_qnames.PatientContext,
                             pm_qnames.LocationContext,
                             pm_qnames.EnsembleContext,
                             pm_qnames.OperatorContext,
                             pm_qnames.WorkflowContext,
                             pm_qnames.MeansContext)
    _child_descriptor_name_mappings = (
        ChildDescriptorMapping(pm_qnames.PatientContext, (pm_qnames.PatientContextDescriptor,)),
        ChildDescriptorMapping(pm_qnames.LocationContext, (pm_qnames.LocationContextDescriptor,)),
        ChildDescriptorMapping(pm_qnames.EnsembleContext, (pm_qnames.EnsembleContextDescriptor,)),
        ChildDescriptorMapping(pm_qnames.OperatorContext, (pm_qnames.OperatorContextDescriptor,)),
        ChildDescriptorMapping(pm_qnames.WorkflowContext, (pm_qnames.WorkflowContextDescriptor,)),
        ChildDescriptorMapping(pm_qnames.MeansContext, (pm_qnames.MeansContextDescriptor,)),
    )


class AbstractContextDescriptorContainer(AbstractDescriptorContainer):
    """Represents AbstractContextDescriptor in BICEPS."""

    is_context_descriptor = True


class PatientContextDescriptorContainer(AbstractContextDescriptorContainer):
    """Represents PatientContextDescriptor in BICEPS."""

    NODETYPE = pm_qnames.PatientContextDescriptor
    STATE_QNAME = pm_qnames.PatientContextState


class LocationContextDescriptorContainer(AbstractContextDescriptorContainer):
    """Represents LocationContextDescriptor in BICEPS."""

    NODETYPE = pm_qnames.LocationContextDescriptor
    STATE_QNAME = pm_qnames.LocationContextState


class WorkflowContextDescriptorContainer(AbstractContextDescriptorContainer):
    """Represents WorkflowContextDescriptor in BICEPS."""

    NODETYPE = pm_qnames.WorkflowContextDescriptor
    STATE_QNAME = pm_qnames.WorkflowContextState


class OperatorContextDescriptorContainer(AbstractContextDescriptorContainer):
    """Represents OperatorContextDescriptor in BICEPS."""

    NODETYPE = pm_qnames.OperatorContextDescriptor
    STATE_QNAME = pm_qnames.OperatorContextState


class MeansContextDescriptorContainer(AbstractContextDescriptorContainer):
    """Represents MeansContextDescriptor in BICEPS."""

    NODETYPE = pm_qnames.MeansContextDescriptor
    STATE_QNAME = pm_qnames.MeansContextState


class EnsembleContextDescriptorContainer(AbstractContextDescriptorContainer):
    """Represents EnsembleContextDescriptor in BICEPS."""

    NODETYPE = pm_qnames.EnsembleContextDescriptor
    STATE_QNAME = pm_qnames.EnsembleContextState


_classes = inspect.getmembers(sys.modules[__name__],
                              lambda member: inspect.isclass(member) and member.__module__ == __name__)
_classes_with_nodetype = [c[1] for c in _classes if hasattr(c[1], 'NODETYPE') and c[1].NODETYPE is not None]
# make a dictionary from found classes: (Key is NODETYPE, value is the class itself

_name_class_lookup = {c.NODETYPE: c for c in _classes_with_nodetype}

_name_class_xtra_lookup = {
    pm_qnames.Battery: BatteryDescriptorContainer,
    pm_qnames.Mds: MdsDescriptorContainer,
    pm_qnames.Vmd: VmdDescriptorContainer,
    pm_qnames.Sco: ScoDescriptorContainer,
    pm_qnames.Channel: ChannelDescriptorContainer,
    pm_qnames.Clock: ClockDescriptorContainer,
    pm_qnames.SystemContext: SystemContextDescriptorContainer,
    pm_qnames.PatientContext: PatientContextDescriptorContainer,
    pm_qnames.LocationContext: LocationContextDescriptorContainer,
    pm_qnames.WorkflowContext: WorkflowContextDescriptorContainer,
    pm_qnames.OperatorContext: OperatorContextDescriptorContainer,
    pm_qnames.MeansContext: MeansContextDescriptorContainer,
    pm_qnames.EnsembleContext: EnsembleContextDescriptorContainer,
    pm_qnames.AlertSystem: AlertSystemDescriptorContainer,
    pm_qnames.AlertCondition: AlertConditionDescriptorContainer,
    pm_qnames.AlertSignal: AlertSignalDescriptorContainer,
}
_name_class_lookup.update(_name_class_xtra_lookup)


def get_container_class(qname: etree.QName) -> type[AbstractDescriptorContainer]:
    """:param qname: a QName instance"""
    return _name_class_lookup.get(qname)



from __future__ import annotations

from typing import TYPE_CHECKING

from sdc11073.namespaces import PrefixesEnum

from .porttypebase import (
    DPWSPortTypeBase,
    WSDLMessageDescription,
    WSDLOperationBinding,
    mk_wsdl_one_way_operation,
    msg_prefix,
)

if TYPE_CHECKING:
    from sdc11073.mdib.mdibbase import MdibVersionGroup
    from sdc11073.mdib.statecontainers import AbstractStateContainer


class WaveformService(DPWSPortTypeBase):
    port_type_name = PrefixesEnum.SDC.tag('WaveformService')
    WSDLMessageDescriptions = (WSDLMessageDescription('Waveform',
                                                      (f'{msg_prefix}:WaveformStreamReport',)),)
    WSDLOperationBindings = (WSDLOperationBinding('Waveform', None, 'literal'),)

    def add_wsdl_port_type(self, parent_node):
        port_type = self._mk_port_type_node(parent_node, True)
        mk_wsdl_one_way_operation(port_type, operation_name='Waveform')

    def _mk_offered_subscriptions(self):
        # unclear if this is needed, it seems wsdl uses Waveform name, action uses WaveformStream
        return [self._sdc_device.mdib.sdc_definitions.Actions.Waveform]

    def send_realtime_samples_report(self, realtime_sample_states: list[AbstractStateContainer],
                                     mdib_version_group: MdibVersionGroup):
        data_model = self._sdc_definitions.data_model
        subscription_mgr = self.hosting_service.subscriptions_manager
        report = data_model.msg_types.WaveformStream()
        report.set_mdib_version_group(mdib_version_group)
        report.State.extend(realtime_sample_states)
        self._logger.debug('sending real time samples report {}', realtime_sample_states)
        subscription_mgr.send_to_subscribers(report, report.action.value, mdib_version_group)



"""A minimal set of codes, used by role provider."""


class NomenclatureCodes:
    """Codes that are used by included role providers."""

    MDC_OP_SET_ALL_ALARMS_AUDIO_PAUSE = '128284'
    MDC_OP_SET_CANCEL_ALARMS_AUDIO_PAUSE = '128285'
    MDC_OP_SET_TIME_SYNC_REF_SRC = '128505'
    MDC_ACT_SET_TIME_ZONE = '68632'



"""wsdiscovery networking input output."""

from __future__ import annotations

import collections
import dataclasses
import logging
import platform
import queue
import random
import selectors
import socket
import struct
import threading
import time
import traceback
from typing import TYPE_CHECKING

from lxml import etree

from sdc11073 import commlog
from sdc11073.exceptions import ValidationError
from sdc11073.wsdiscovery.common import MULTICAST_IPV4_ADDRESS, message_reader

if TYPE_CHECKING:
    from logging import Logger

    from sdc11073.pysoap.msgfactory import CreatedMessage
    from sdc11073.wsdiscovery.wsdimpl import WSDiscovery

BUFFER_SIZE = 0xffff


@dataclasses.dataclass(frozen=True)
class _UdpRepeatParams:
    """Udp messages are send multiple times with random gaps. These parameters define the limits of the randomness."""

    max_initial_delay_ms: int  # max. delay before sending the fist datagram
    repeat: int  # number of repeated sends
    min_delay_ms: int  # minimum delay for first repetition in ms
    max_delay_ms: int  # maximal delay for first repetition in ms
    upper_delay_ms: int  # max. delay between repetitions in ms
    # (gap is doubled for each further repetition, but not more than this value)


UNICAST_REPEAT_PARAMS = _UdpRepeatParams(500, 2, 50, 250, 500)
MULTICAST_REPEAT_PARAMS = _UdpRepeatParams(500, 4, 50, 250, 500)

# these time constants control the send-loop
SEND_LOOP_IDLE_SLEEP = 0.1
SEND_LOOP_BUSY_SLEEP = 0.01


@dataclasses.dataclass(frozen=True)
class OutgoingMessage:
    """OutgoingMessage instances contain a soap envelope, destination address and multicast / unicast information."""

    created_message: CreatedMessage
    addr: str
    port: int

    def __repr__(self):
        return (f"{self.__class__.__name__}(addr={self.addr}, port={self.port}, "
                f"created_message={self.created_message.serialize()})")


class NetworkingThread:
    """Has one thread for sending and one for receiving."""

    @dataclasses.dataclass(order=True)
    class _EnqueuedMessage:
        send_time: float
        msg: OutgoingMessage = dataclasses.field(compare=False)
        repeat: int

    def __init__(self,
                 my_ip_address: str,
                 wsd: WSDiscovery,
                 logger: Logger,
                 multicast_port: int,
                 multicast_ttl: int):
        self._my_ip_address = my_ip_address
        self._wsd = wsd
        self._logger = logger
        self.multicast_port = multicast_port
        self._recv_thread = None
        self._qread_thread = None
        self._send_thread = None
        self._quit_recv_event = threading.Event()
        self._quit_send_event = threading.Event()
        self._send_queue = queue.PriorityQueue(10000)
        self._read_queue = queue.Queue(10000)
        self._known_message_ids = collections.deque(maxlen=200)
        self._inbound_selector = selectors.DefaultSelector()
        self._outbound_selector = selectors.DefaultSelector()
        self.multi_in = self._create_multicast_in_socket(my_ip_address, multicast_port)
        self.multi_out_uni_in_out = self._create_multi_out_uni_in_out_socket(my_ip_address, multicast_ttl)

    def _register_inbound_socket(self, sock: socket.SocketType):
        self._inbound_selector.register(sock, selectors.EVENT_READ)
        self._logger.info('registered inbound socket on %s:%d', *sock.getsockname())

    def _register_outbound_socket(self, sock: socket.SocketType):
        self._outbound_selector.register(sock, selectors.EVENT_WRITE)
        self._logger.info('registered outbound socket on %s:%d', *sock.getsockname())

    def _create_multi_out_uni_in_out_socket(self, addr: str, multicast_ttl: int) -> socket.SocketType:
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, multicast_ttl)
        sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_IF, socket.inet_aton(addr))
        # set port explicitly when creating (it would otherwise be set after sending first message via this socket)
        sock.bind((addr, 0))
        self._register_outbound_socket(sock)
        self._register_inbound_socket(sock)
        return sock

    def _create_multicast_in_socket(self, addr: str, port: int) -> socket.SocketType:
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        if platform.system() != 'Windows':
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
            sock.bind((MULTICAST_IPV4_ADDRESS, port))
        else:
            sock.bind((addr, port))
        sock.setblocking(False)
        _addr = struct.pack("4s4s", socket.inet_aton(MULTICAST_IPV4_ADDRESS), socket.inet_aton(addr))
        sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, _addr)
        self._register_inbound_socket(sock)
        return sock

    def add_outbound_message(self, msg: CreatedMessage, addr: str, port: int, repeat_params: _UdpRepeatParams):
        """Add a message to the sending queue."""
        self._logger.debug('adding outbound message with Id "%s" to sending queue',
                           msg.p_msg.header_info_block.MessageID)
        self._known_message_ids.appendleft(msg.p_msg.header_info_block.MessageID)
        self._repeated_enqueue_msg(OutgoingMessage(msg, addr, port), repeat_params)

    def _repeated_enqueue_msg(self, msg: OutgoingMessage, delay_params: _UdpRepeatParams):
        if self._quit_send_event.is_set():
            self._logger.warning('_repeated_enqueue_msg: sending thread not running - message will be dropped - %s',
                                 msg)
            return
        initial_delay_ms = random.randint(0, delay_params.max_initial_delay_ms)
        next_send = time.time() + initial_delay_ms / 1000.0
        delta_t = random.randrange(delay_params.min_delay_ms, delay_params.max_delay_ms) / 1000.0  # millisec -> seconds
        self._send_queue.put(self._EnqueuedMessage(next_send, msg, 1))
        for i in range(delay_params.repeat):
            next_send += delta_t
            self._send_queue.put(self._EnqueuedMessage(next_send, msg, i + 2))
            delta_t = min(delta_t * 2, delay_params.upper_delay_ms)

    def _run_send(self):
        """send-loop."""
        while not self._quit_send_event.is_set() or not self._send_queue.empty():
            if self._send_queue.empty():
                time.sleep(SEND_LOOP_IDLE_SLEEP)  # nothing to do currently
                continue
            if self._send_queue.queue[0].send_time <= time.time():
                enqueued_msg = self._send_queue.get()
                for key, _ in self._outbound_selector.select(timeout=0.1):
                    self._send_msg(enqueued_msg, key.fileobj)
            else:
                time.sleep(SEND_LOOP_BUSY_SLEEP)  # this creates a 10ms raster for sending, but that is good enough

    def _run_recv(self):
        """Run by thread."""
        while not self._quit_recv_event.is_set():
            try:
                self._recv_messages()
            except:  # noqa: E722, PERF203 use bare except here, this is a catch-all that keeps thread running.
                self._logger.exception('exception during receiving')

    def _recv_messages(self):
        """For performance reasons this thread only writes to a queue, no parsing etc."""
        for key, _ in self._inbound_selector.select(timeout=0.1):
            sock: socket.SocketType = key.fileobj
            try:
                data, addr = sock.recvfrom(BUFFER_SIZE)
            except OSError as exc:
                self._logger.warning('socket read error %s', exc)
                time.sleep(0.01)
                continue
            self._logger.debug('received data on my socket %s', sock.getsockname())
            self._add_to_recv_queue(addr, data)

    def _add_to_recv_queue(self, addr: str, data: bytes):
        # method is needed for testing
        self._read_queue.put((addr, data))

    def _run_q_read(self):
        """Read from internal queue and process message."""
        while not self._quit_recv_event.is_set():
            try:
                incoming = self._read_queue.get(timeout=0.1)
            except queue.Empty:  # noqa: PERF203
                pass
            else:
                addr, data = incoming
                if b"http://schemas.xmlsoap.org/ws/2005/04/discovery" in data:
                    continue  # older version of discovery standard, ignore completely.
                logging.getLogger(commlog.DISCOVERY_IN).debug(data, extra={'ip_address': addr[0]})
                try:
                    try:
                        received_message = message_reader.read_received_message(data, validate=True)
                    except (etree.XMLSyntaxError, ValidationError) as ex:
                        self._logger.info('_run_q_read: received invalid message from %r, ignoring it (error=%s)', addr,
                                          ex)
                    else:
                        mid = received_message.p_msg.header_info_block.MessageID
                        if mid in self._known_message_ids:
                            self._logger.debug('incoming message already known: %s (from %r, Id %s).',
                                               received_message.action, addr, mid)
                            continue
                        self._logger.debug('new incoming message: %s (from %r, Id %s).',
                                           received_message.action, addr, mid)
                        self._known_message_ids.appendleft(mid)
                        self._wsd.handle_received_message(received_message, addr)
                except Exception:  # noqa: BLE001
                    self._logger.error('_run_q_read: %s', traceback.format_exc())  # noqa: TRY400

    def _send_msg(self, q_msg: _EnqueuedMessage, s: socket.socket):
        msg = q_msg.msg
        data = msg.created_message.serialize()
        self._logger.debug('send message %d bytes (%d) action=%s: to=%s:%r id=%s',
                           len(data),
                           q_msg.repeat,
                           msg.created_message.p_msg.header_info_block.Action,
                           msg.addr, msg.port,
                           msg.created_message.p_msg.header_info_block.MessageID)
        try:
            s.sendto(data, (msg.addr, msg.port))
        except:  # noqa: E722 use bare except here, this is a catch-all that keeps thread running.
            self._logger.exception('exception during sending data to %s:%d: %s', msg.addr, msg.port, data)
        else:
            # log this if there was no exception during send
            logging.getLogger(commlog.DISCOVERY_OUT).debug(data, extra={'ip_address': msg.addr})

    def start(self):
        """Start working for the sending and receiving queue."""
        self._logger.debug('%s: starting ', self.__class__.__name__)
        self._recv_thread = threading.Thread(target=self._run_recv, name='wsd.recvThread', daemon=True)
        self._qread_thread = threading.Thread(target=self._run_q_read, name='wsd.qreadThread', daemon=True)
        self._send_thread = threading.Thread(target=self._run_send, name='wsd.sendThread', daemon=True)
        self._recv_thread.start()
        self._qread_thread.start()
        self._send_thread.start()

    def schedule_stop(self):
        """Schedule stopping the thread.

        Use join() to wait, until thread really has been stopped.
        """
        self._logger.debug('%s: schedule_stop ', self.__class__.__name__)
        self._quit_recv_event.set()
        self._quit_send_event.set()

    def join(self):
        """Join threads and close sockets."""
        self._logger.debug('%s: join... ', self.__class__.__name__)
        self._recv_thread.join()
        self._send_thread.join()
        self._qread_thread.join()
        self._recv_thread = None
        self._send_thread = None
        self._qread_thread = None
        self.multi_in.close()
        self.multi_out_uni_in_out.close()
        self._inbound_selector.close()
        self._outbound_selector.close()
        self._logger.debug('%s: ... join done', self.__class__.__name__)



from __future__ import annotations

from typing import TYPE_CHECKING

from sdc11073.namespaces import PrefixesEnum
from sdc11073.xml_types.addressing_types import HeaderInformationBlock

from .serviceclientbase import GetRequestResult, HostedServiceClient

if TYPE_CHECKING:
    from sdc11073.consumer.manipulator import RequestManipulatorProtocol
    from collections.abc import Iterable

class CTreeServiceClient(HostedServiceClient):
    """Client for ContainmentTreeService."""

    port_type_name = PrefixesEnum.SDC.tag('ContainmentTreeService')

    def get_descriptor(self, handles: Iterable[str],
                       request_manipulator: RequestManipulatorProtocol | None = None) -> GetRequestResult:
        """Send a GetDescriptor request to provider."""
        data_model = self._sdc_definitions.data_model
        request = data_model.msg_types.GetDescriptor()
        request.HandleRef.extend(handles)
        inf = HeaderInformationBlock(action=request.action, addr_to=self.endpoint_reference.Address)
        message = self._msg_factory.mk_soap_message(inf, payload=request)

        received_message_data = self.post_message(message, request_manipulator=request_manipulator)
        cls = data_model.msg_types.GetDescriptorResponse
        report = cls.from_node(received_message_data.p_msg.msg_node)
        return GetRequestResult(received_message_data, report)

    def get_containment_tree(self, handles: Iterable[str],
                             request_manipulator: RequestManipulatorProtocol | None = None) -> GetRequestResult:
        """Send a GetContainmentTree request to provider."""
        data_model = self._sdc_definitions.data_model
        request = data_model.msg_types.GetContainmentTree()
        request.HandleRef.extend(handles)
        inf = HeaderInformationBlock(action=request.action, addr_to=self.endpoint_reference.Address)
        message = self._msg_factory.mk_soap_message(inf, payload=request)

        received_message_data = self.post_message(message, request_manipulator=request_manipulator)
        cls = data_model.msg_types.GetContainmentTreeResponse
        report = cls.from_node(received_message_data.p_msg.msg_node)
        return GetRequestResult(received_message_data, report)



"""The module contains the implementation of MdibBase plus entity interface."""

from __future__ import annotations

import copy
import traceback
import uuid
from dataclasses import dataclass
from threading import Lock
from typing import TYPE_CHECKING, Any

from lxml import etree

from sdc11073 import multikey
from sdc11073 import observableproperties as properties
from sdc11073.etc import apply_map
from sdc11073.xml_types.pm_types import Coding, have_matching_codes

if TYPE_CHECKING:
    from collections.abc import Iterable

    from lxml.etree import QName

    from sdc11073 import xml_utils
    from sdc11073.definitions_base import BaseDefinitions
    from sdc11073.loghelper import LoggerAdapter
    from sdc11073.xml_types.pm_types import CodedValue

    from .descriptorcontainers import AbstractDescriptorContainer, AbstractOperationDescriptorContainer
    from .statecontainers import AbstractMultiStateContainer, AbstractStateContainer


@dataclass
class MdibVersionGroup:
    """These 3 values define an mdib version."""

    mdib_version: int
    sequence_id: str
    instance_id: int | None


class _MultikeyWithVersionLookup(multikey.MultiKeyLookup):
    """_MultikeyWithVersionLookup keeps track of versions of removed objects.

    If a descriptor od state gets removed from mdib and later is added again, the version of that instance
    must be greater than the last version before it was removed.
    """

    def __init__(self):
        super().__init__()
        self.handle_version_lookup = {}

    def remove_object(self, obj: Any):
        if obj is not None:
            self._save_version(obj)
        super().remove_object(obj)

    def remove_object_no_lock(self, obj: Any):
        """Remove object from table without locking."""
        if obj is not None:
            self._save_version(obj)
        super().remove_object_no_lock(obj)

    def remove_objects_no_lock(self, objects: list[Any]):
        apply_map(self._save_version, [obj for obj in objects if obj is not None])
        super().remove_objects_no_lock(objects)


class DescriptorsLookup(_MultikeyWithVersionLookup):
    """DescriptorsLookup is the table-like storage for descriptors.

    It has the following search indices:
     - handle is the index for descriptor.Handle.
     - parent_handle is the index for descriptor.parent_handle, it finds all children of a queried descriptor.
     - NODETYPE is the index for descriptor.NODETYPE. It finds all children of a queried type.
       This index works only for exact matches, class hierarchy is unknown here. E.g. AlertDescriptor only returns
       AlertDescriptor objects, not LimitAlertDescriptor!
     - coding is the index for descriptor.coding.
     - condition_signaled is the index for descriptor.ConditionSignaled, it finds only AlertSignalDescriptors.
     - source is the index for descriptor.Source, it finds only AlertConditionDescriptors.
    """

    handle: multikey.UIndexDefinition[str, list[AbstractDescriptorContainer]]
    parent_handle: multikey.IndexDefinition[str, list[AbstractDescriptorContainer]]
    NODETYPE: multikey.IndexDefinition[etree.QName, list[AbstractDescriptorContainer]]
    coding: multikey.IndexDefinition[Coding, list[AbstractDescriptorContainer]]
    condition_signaled: multikey.IndexDefinition[str, list[AbstractDescriptorContainer]]
    source: multikey.IndexDefinition[str, list[AbstractDescriptorContainer]]

    def __init__(self):
        super().__init__()
        self.add_index('handle', multikey.UIndexDefinition(lambda obj: obj.Handle))
        self.add_index('parent_handle', multikey.IndexDefinition(lambda obj: obj.parent_handle))
        self.add_index('NODETYPE', multikey.IndexDefinition(lambda obj: obj.NODETYPE))
        self.add_index('coding', multikey.IndexDefinition(lambda obj: obj.coding))
        self.add_index(
            'condition_signaled',
            multikey.IndexDefinition(lambda obj: obj.ConditionSignaled, index_none_values=False),
        )
        # an index to find all alert conditions for a metric (AlertCondition is the only class that has a
        # "Source" attribute, therefore this simple approach without type testing is sufficient):
        self.add_index('source', multikey.IndexDefinition1n(lambda obj: obj.Source, index_none_values=False))

    def _save_version(self, obj: AbstractDescriptorContainer):
        self.handle_version_lookup[obj.Handle] = obj.DescriptorVersion

    def set_version(self, obj: AbstractDescriptorContainer):
        """Set DescriptorVersion of obj if descriptor with same handle existed before."""
        version = self.handle_version_lookup.get(obj.Handle)
        if version is not None:
            obj.DescriptorVersion = version + 1

    def add_object(self, obj: AbstractDescriptorContainer):
        """Append object with locking."""
        with self._lock:
            self.add_object_no_lock(obj)

    def add_object_no_lock(self, obj: AbstractDescriptorContainer):
        """Append object without locking."""
        super().add_object_no_lock(obj)

    def add_objects(self, objects: list[AbstractDescriptorContainer]):
        """Append objects with locking."""
        with self._lock:
            self.add_objects_no_lock(objects)

    def add_objects_no_lock(self, objects: list[AbstractDescriptorContainer]):
        """Append objects without locking."""
        apply_map(self.add_object_no_lock, objects)

    def remove_object(self, obj: AbstractDescriptorContainer):
        """Remove object from table."""
        keys = self._object_ids.get(id(obj))
        if keys is None:
            return
        with self._lock:
            self.remove_object_no_lock(obj)

    def remove_object_no_lock(self, obj: AbstractDescriptorContainer):
        """Remove object from table without locking."""
        super().remove_object_no_lock(obj)

    def remove_objects(self, objects: list[AbstractDescriptorContainer]):
        """Remove objects from table with locking."""
        with self._lock:
            self.remove_objects_no_lock(objects)

    def remove_objects_no_lock(self, objects: list[AbstractDescriptorContainer]):
        """Remove objects from table without locking."""
        apply_map(self.remove_object_no_lock, objects)


class StatesLookup(_MultikeyWithVersionLookup):
    """StatesLookup is the table-like storage for states.

    It has search indices:
     - descriptor_handle is the index for descriptor.DescriptorHandle.
     - NODETYPE is the index for descriptor.NODETYPE. It finds all children of a queried type.
       This index works only for exact matches, class hierarchy is unknown here. E.g. AlertState only returns
       AlertState objects, not LimitAlertState!
    """

    descriptor_handle: multikey.UIndexDefinition[str, list[AbstractDescriptorContainer]]
    NODETYPE: multikey.IndexDefinition[etree.QName, list[AbstractDescriptorContainer]]

    def __init__(self):
        super().__init__()
        self.add_index('descriptor_handle', multikey.UIndexDefinition(lambda obj: obj.DescriptorHandle))
        self.add_index('NODETYPE', multikey.IndexDefinition(lambda obj: obj.NODETYPE, index_none_values=False))

    def _save_version(self, obj: AbstractStateContainer):
        self.handle_version_lookup[obj.DescriptorHandle] = obj.StateVersion

    def set_version(self, obj: AbstractStateContainer):
        """Set StateVersion of obj if state with same handle existed before."""
        version = self.handle_version_lookup.get(obj.DescriptorHandle)
        if version is not None:
            obj.StateVersion = version + 1

    def add_object_no_lock(self, obj: AbstractStateContainer):
        """Append object without locking."""
        if obj.is_multi_state:
            raise ValueError('no MultiState!')
        super().add_object_no_lock(obj)


class MultiStatesLookup(_MultikeyWithVersionLookup):
    """MultiStatesLookup is the table-like storage for multi-states.

    It has search indices:
     - descriptor_handle is the index for descriptor.DescriptorHandle.
     - handle is the index for descriptor.Handle.
     - NODETYPE is the index for descriptor.NODETYPE. It finds all children of a queried type.
       This index works only for exact matches, class hierarchy is unknown here.
       AlertDescriptor objects, not LimitAlertDescriptor!
    """

    descriptor_handle: multikey.IndexDefinition[str, list[AbstractMultiStateContainer]]
    handle: multikey.UIndexDefinition[str, list[AbstractMultiStateContainer]]
    NODETYPE: multikey.IndexDefinition[etree.QName, list[AbstractMultiStateContainer]]

    def __init__(self):
        super().__init__()
        self.add_index('descriptor_handle', multikey.IndexDefinition(lambda obj: obj.DescriptorHandle))
        self.add_index('handle', multikey.UIndexDefinition(lambda obj: obj.Handle, index_none_values=False))
        self.add_index('NODETYPE', multikey.IndexDefinition(lambda obj: obj.NODETYPE, index_none_values=False))

    def _save_version(self, obj: AbstractMultiStateContainer):
        self.handle_version_lookup[obj.Handle] = obj.StateVersion

    def set_version(self, obj: AbstractMultiStateContainer):
        """Set StateVersion of obj if state with same handle existed before."""
        version = self.handle_version_lookup.get(obj.Handle)
        if version is not None:
            obj.StateVersion = version + 1


class _EntityBase:
    def __init__(self, mdib: MdibBase, descriptor: AbstractDescriptorContainer):
        self._mdib = mdib
        self.descriptor = descriptor

    @property
    def handle(self) -> str:
        return self.descriptor.Handle

    @property
    def parent_handle(self) -> str:
        return self.descriptor.parent_handle

    @property
    def node_type(self) -> QName:
        return self.descriptor.NODETYPE

    def update(self):
        """Update the entity from current data in mdib."""
        orig = self._mdib.descriptions.handle.get_one(self.handle)
        self.descriptor.update_from_other_container(orig)


class Entity(_EntityBase):
    """Groups descriptor and state."""

    def __init__(self, mdib: MdibBase, descriptor: AbstractDescriptorContainer, state: AbstractStateContainer):
        super().__init__(mdib, descriptor)
        self.state = state

    @property
    def is_multi_state(self) -> bool:
        """Return False because this is not a multi state entity."""
        return False

    def update(self):
        """Update the entity from current data in mdib."""
        super().update()
        orig = self._mdib.states.get_one(self.handle)
        self.state.update_from_other_container(orig)


class MultiStateEntity(_EntityBase):
    """Groups descriptor and list of multi-states."""

    def __init__(
        self,
        mdib: MdibBase,
        descriptor: AbstractDescriptorContainer,
        states: list[AbstractMultiStateContainer],
    ):
        super().__init__(mdib, descriptor)
        self.states: dict[str, AbstractMultiStateContainer] = {s.Handle: s for s in states}

    @property
    def is_multi_state(self) -> bool:
        """Return True because this is a multi state entity."""
        return True

    def update(self):
        """Update the entity from current data in mdib."""
        super().update()

        all_orig_states = self._mdib.context_states.descriptor_handle.get(self.handle, [])
        states_dict = {st.Handle: st for st in all_orig_states}
        # update existing states, remove deleted ones
        for state in list(self.states.values()):
            orig = states_dict.get(state.Handle)
            if orig is not None:
                state.update_from_other_container(orig)
            else:
                self.states.pop(state.Handle)
        # add new states
        for handle, _ in states_dict.items():  # noqa: PERF102
            if handle not in self.states:
                self.states[handle] = states_dict[handle].mk_copy()

    def new_state(self, state_handle: str | None = None) -> AbstractMultiStateContainer:
        """Create a new state."""
        if state_handle in self.states:
            msg = f'State handle {state_handle} already exists in {self.__class__.__name__}, handle = {self.handle}'
            raise ValueError(msg)
        cls = self._mdib.data_model.get_state_container_class(self.descriptor.STATE_QNAME)
        state = cls(descriptor_container=self.descriptor)
        state.Handle = state_handle or uuid.uuid4().hex
        self.states[state.Handle] = state
        return state


class EntityGetter:
    """Implementation of EntityGetterProtocol for MdibBase."""

    def __init__(self, mdib: MdibBase):
        self._mdib = mdib

    def by_handle(self, handle: str) -> Entity | MultiStateEntity | None:
        """Return entity with given handle."""
        descriptor = self._mdib.descriptions.handle.get_one(handle, allow_none=True)
        if descriptor is None:
            return None
        return self._mk_entity(descriptor)

    def by_node_type(self, node_type: QName) -> list[Entity | MultiStateEntity]:
        """Return all entities with given node type."""
        descriptors = self._mdib.descriptions.NODETYPE.get(node_type, [])
        return [self._mk_entity(d) for d in descriptors]

    def by_parent_handle(self, parent_handle: str | None) -> list[Entity | MultiStateEntity]:
        """Return all entities with descriptors parent_handle == provided parent_handle."""
        descriptors = self._mdib.descriptions.parent_handle.get(parent_handle, [])
        return [self._mk_entity(d) for d in descriptors]

    def by_coding(self, coding: Coding) -> list[Entity | MultiStateEntity]:
        """Return all entities with descriptors type are equivalent to codeding."""
        descriptors = [
            d for d in self._mdib.descriptions.objects if d.Type is not None and d.Type.is_equivalent(coding)
        ]
        return [self._mk_entity(d) for d in descriptors]

    def by_coded_value(self, coded_value: CodedValue) -> list[Entity | MultiStateEntity]:
        """Return all entities with descriptors type are equivalent to coded_value."""
        descriptors = [d for d in self._mdib.descriptions.objects if d.Type.is_equivalent(coded_value)]
        return [self._mk_entity(d) for d in descriptors]

    def _mk_entity(self, descriptor: AbstractDescriptorContainer) -> Entity | MultiStateEntity:
        if descriptor.is_context_descriptor:
            states = self._mdib.context_states.descriptor_handle.get(descriptor.Handle, [])
            return MultiStateEntity(self._mdib, copy.deepcopy(descriptor), copy.deepcopy(states))
        state = self._mdib.states.descriptor_handle.get_one(descriptor.Handle)
        return Entity(self._mdib, copy.deepcopy(descriptor), copy.deepcopy(state))

    def items(self) -> Iterable[tuple[str, Entity | MultiStateEntity]]:
        """Return the items of a dictionary."""
        for descriptor in self._mdib.descriptions.objects:
            yield descriptor.Handle, self._mk_entity(descriptor)

    def __len__(self) -> int:
        """Return number of entities."""
        return len(self._mdib.descriptions.objects)


class MdibBase:
    """Base class with common functionality of provider mdib and consumer mdib."""

    # these observables can be used to watch any change of data in the mdib.
    # They contain lists of containers that were changed.
    # every transaction (device mdib) or notification (client mdib) will report their changes here.
    metrics_by_handle = properties.ObservableProperty(fire_only_on_changed_value=False)
    waveform_by_handle = properties.ObservableProperty(fire_only_on_changed_value=False)
    alert_by_handle = properties.ObservableProperty(fire_only_on_changed_value=False)
    context_by_handle = properties.ObservableProperty(fire_only_on_changed_value=False)
    component_by_handle = properties.ObservableProperty(fire_only_on_changed_value=False)
    new_descriptors_by_handle = properties.ObservableProperty(fire_only_on_changed_value=False)
    updated_descriptors_by_handle = properties.ObservableProperty(fire_only_on_changed_value=False)
    deleted_descriptors_by_handle = properties.ObservableProperty(fire_only_on_changed_value=False)
    deleted_states_by_handle = properties.ObservableProperty(
        fire_only_on_changed_value=False,
    )  # is a result of deleted descriptors
    description_modifications = properties.ObservableProperty(fire_only_on_changed_value=False)
    operation_by_handle = properties.ObservableProperty(fire_only_on_changed_value=False)
    sequence_id = properties.ObservableProperty()
    instance_id = properties.ObservableProperty()

    def __init__(self, sdc_definitions: type[BaseDefinitions], logger: LoggerAdapter):
        """Construct MdibBase.

        :param sdc_definitions: a class derived from BaseDefinitions
        """
        self.sdc_definitions = sdc_definitions
        self.data_model = sdc_definitions.data_model
        self._logger = logger
        self.nsmapper = sdc_definitions.data_model.ns_helper
        self.mdib_version = 0
        self.sequence_id = ''  # needs to be set to a reasonable value by derived class
        self.instance_id = None  # None or an unsigned int
        self.log_prefix = ''
        self.descriptions = DescriptorsLookup()
        self.states = StatesLookup()
        self.context_states = MultiStatesLookup()
        self.mdib_lock = Lock()
        self.mdstate_version = 0
        self.mddescription_version = 0

    @property
    def logger(self) -> LoggerAdapter:
        """Return the logger."""
        return self._logger

    @property
    def mdib_version_group(self) -> MdibVersionGroup:
        """Get current version data."""
        return MdibVersionGroup(self.mdib_version, self.sequence_id, self.instance_id)

    def add_description_containers(self, description_containers: list[AbstractDescriptorContainer]):
        """Initialize descriptions member with provided descriptors.

        :param description_containers: a list of DescriptorContainer objects
        """
        new_descriptor_by_handle = {}
        with self.descriptions.lock:
            for description_container in description_containers:
                self.descriptions.add_object_no_lock(description_container)
                new_descriptor_by_handle[description_container.Handle] = description_container

        # finally update observable property
        if new_descriptor_by_handle:
            self.new_descriptors_by_handle = new_descriptor_by_handle

    def clear_states(self):
        """Remove all states and context states."""
        with self.states.lock:
            self.states.clear()
            self.context_states.clear()

        # clear also the observable properties
        self.metrics_by_handle = None
        self.waveform_by_handle = None
        self.alert_by_handle = None
        self.context_by_handle = None
        self.component_by_handle = None
        self.operation_by_handle = None

    def _set_descriptor_container_reference(self, state_container: AbstractStateContainer):
        """Set state_container.descriptor_container if all is fine, otherwise logs error."""
        descriptor_container = self.descriptions.handle.get_one(state_container.DescriptorHandle, allow_none=True)
        if descriptor_container is None:
            self._logger.warning(  # noqa: PLE1205
                'state "{}" (type={}) has no descriptor in mdib!',
                state_container.DescriptorHandle,
                state_container.NODETYPE,
            )
        elif descriptor_container.DescriptorVersion == state_container.DescriptorVersion:
            state_container.descriptor_container = descriptor_container
        else:
            self._logger.warning(  # noqa: PLE1205
                'state "{}" (type={}) : descriptor version expect "{}", found "{}"',
                state_container.DescriptorHandle,
                state_container.NODETYPE,
                descriptor_container.DescriptorVersion,
                state_container.DescriptorVersion,
            )

    def add_state_containers(self, state_containers: list[AbstractStateContainer | AbstractMultiStateContainer]):
        """Add states to self.states and self.context_states.

        This method does not update the observables, so use with care!
        :param state_containers: a list of StateContainer objects.
        """
        for state_container in state_containers:
            if state_container.descriptor_container is None:
                self._set_descriptor_container_reference(state_container)
            my_multikey = self.context_states if state_container.is_context_state else self.states
            try:
                my_multikey.add_object(state_container)
            except KeyError as ex:
                if state_container.is_context_state:
                    self._logger.error(  # noqa: PLE1205 TRY400
                        'add_state_containers: {}, Handle={}; {}',
                        ex,
                        state_container.Handle,
                        traceback.format_exc(),
                    )
                else:
                    self._logger.error(  # noqa: PLE1205 TRY400
                        'add_state_containers: {}, DescriptorHandle={}; {}',
                        ex,
                        state_container.DescriptorHandle,
                        traceback.format_exc(),
                    )

    def _reconstruct_md_description(self) -> xml_utils.LxmlElement:
        """Build dom tree of descriptors from current data."""
        pm = self.data_model.pm_names
        doc_nsmap = self.nsmapper.ns_map
        root_containers = self.descriptions.parent_handle.get(None) or []
        md_description_node = etree.Element(
            pm.MdDescription,
            attrib={'DescriptionVersion': str(self.mddescription_version)},
            nsmap=doc_nsmap,
        )
        for root_container in root_containers:
            self.make_descriptor_node(root_container, md_description_node, tag=pm.Mds, set_xsi_type=False)
        return md_description_node

    def make_descriptor_node(
        self,
        descriptor_container: AbstractDescriptorContainer,
        parent_node: xml_utils.LxmlElement,
        tag: etree.QName,
        set_xsi_type: bool = True,
    ) -> xml_utils.LxmlElement:
        """Create a lxml etree node with subtree from instance data.

        :param descriptor_container: a descriptor container instance
        :param parent_node: parent node
        :param tag: tag of node
        :param set_xsi_type: if true, the NODETYPE will be used to set the xsi:type attribute of the node
        :return: an etree node.
        """
        ns_map = (
            self.nsmapper.partial_map(self.nsmapper.PM, self.nsmapper.XSI)
            if set_xsi_type
            else self.nsmapper.partial_map(self.nsmapper.PM)
        )
        node = etree.SubElement(parent_node, tag, attrib={'Handle': descriptor_container.Handle}, nsmap=ns_map)
        descriptor_container.update_node(node, self.nsmapper, set_xsi_type)  # create all
        child_list = self.descriptions.parent_handle.get(descriptor_container.Handle, [])
        # append all child containers, then bring all child elements in correct order
        for child in child_list:
            child_tag, set_xsi = descriptor_container.tag_name_for_child_descriptor(child.NODETYPE)
            self.make_descriptor_node(child, node, child_tag, set_xsi)
        descriptor_container.sort_child_nodes(node)
        return node

    def _reconstruct_mdib(self, add_context_states: bool) -> xml_utils.LxmlElement:
        """Build dom tree of mdib from current data.

        If add_context_states is False, context states are not included.
        """
        pm = self.data_model.pm_names
        msg = self.data_model.msg_names
        doc_nsmap = self.nsmapper.ns_map
        mdib_node = etree.Element(msg.Mdib, nsmap=doc_nsmap)
        mdib_node.set('MdibVersion', str(self.mdib_version))
        mdib_node.set('SequenceId', self.sequence_id)
        if self.instance_id is not None:
            mdib_node.set('InstanceId', str(self.instance_id))
        md_description_node = self._reconstruct_md_description()
        mdib_node.append(md_description_node)

        # add a list of states
        md_state_node = etree.SubElement(
            mdib_node,
            pm.MdState,
            attrib={'StateVersion': str(self.mdstate_version)},
            nsmap=doc_nsmap,
        )
        tag = pm.State
        for state_container in self.states.objects:
            md_state_node.append(state_container.mk_state_node(tag, self.nsmapper))
        if add_context_states:
            for state_container in self.context_states.objects:
                md_state_node.append(state_container.mk_state_node(tag, self.nsmapper))
        return mdib_node

    def reconstruct_md_description(self) -> (xml_utils.LxmlElement, MdibVersionGroup):
        """Build dom tree of descriptors from current data."""
        with self.mdib_lock:
            node = self._reconstruct_md_description()
            return node, self.mdib_version_group

    def reconstruct_mdib(self) -> (xml_utils.LxmlElement, MdibVersionGroup):
        """Build dom tree from current data.

        This method does not include context states!
        """
        with self.mdib_lock:
            return self._reconstruct_mdib(add_context_states=False), self.mdib_version_group

    def reconstruct_mdib_with_context_states(self) -> (xml_utils.LxmlElement, MdibVersionGroup):
        """Build dom tree from current data.

        This method includes the context states.
        """
        with self.mdib_lock:
            return self._reconstruct_mdib(add_context_states=True), self.mdib_version_group

    def _get_child_descriptors_by_code(self, parent_handle: str, code: Coding) -> list[AbstractDescriptorContainer]:
        descriptors = self.descriptions.parent_handle.get(parent_handle, [])
        if len(descriptors) == 0:
            return []
        with_types = [d for d in descriptors if d.Type is not None]
        return [d for d in with_types if have_matching_codes(d.Type, code)]

    def get_metric_descriptor_by_code(
        self,
        vmd_code: [Coding, CodedValue],
        channel_code: [Coding, CodedValue],
        metric_code: [Coding, CodedValue],
    ) -> AbstractDescriptorContainer | None:
        """get_metric_descriptor_by_code is the "correct" way to find a descriptor.

        Using handles is shaky, because they have no meaning and can change over time!
        """
        pm = self.data_model.pm_names
        all_vmds = self.descriptions.NODETYPE.get(pm.VmdDescriptor, [])
        matching_vmd_list = [d for d in all_vmds if have_matching_codes(d.Type, vmd_code)]
        for vmd in matching_vmd_list:
            matching_channels = self._get_child_descriptors_by_code(vmd.Handle, channel_code)
            for channel in matching_channels:
                matching_metrics = self._get_child_descriptors_by_code(channel.Handle, metric_code)
                if len(matching_metrics) == 1:
                    return matching_metrics[0]
                if len(matching_metrics) > 1:
                    msg = f'found multiple metrics for vmd={vmd_code} channel={channel_code} metric={metric_code}'
                    raise ValueError(msg)
        return None

    def get_operations_for_metric(
        self,
        vmd_code: [Coding, CodedValue],
        channel_code: [Coding, CodedValue],
        metric_code: [Coding, CodedValue],
    ) -> list[AbstractDescriptorContainer]:
        """get_operations_for_metric is the "correct" way to find an operation.

        Using handles is shaky, because they have no meaning and can change over time!
        """
        descriptor_container = self.get_metric_descriptor_by_code(vmd_code, channel_code, metric_code)
        return self.get_operation_descriptors_for_descriptor_handle(descriptor_container.Handle)

    def get_operation_descriptors_for_descriptor_handle(
        self,
        descriptor_handle: str,
        **additional_filters: Any,
    ) -> list[AbstractDescriptorContainer]:
        """Get operation descriptors that have descriptor_handle as OperationTarget.

        :param descriptor_handle: the handle for that operations shall be found
        :return: a list with operation descriptors that have descriptor_handle as OperationTarget. List can be empty
        :additionalFilters: optional filters for the key = name of member attribute, value = expected value
            example: NODETYPE=pm.SetContextStateOperationDescriptor filters for SetContextStateOperation descriptors
        """
        all_operation_containers = self.get_operation_descriptors()
        my_operations = [op_c for op_c in all_operation_containers if op_c.OperationTarget == descriptor_handle]
        for key, value in additional_filters.items():
            my_operations = [op for op in my_operations if getattr(op, key) == value]
        return my_operations

    def get_operation_descriptors(self) -> list[AbstractOperationDescriptorContainer]:
        """Get a list of all operation descriptors."""
        pm = self.data_model.pm_names
        result = []
        for node_type in (
            pm.SetValueOperationDescriptor,
            pm.SetStringOperationDescriptor,
            pm.ActivateOperationDescriptor,
            pm.SetContextStateOperationDescriptor,
            pm.SetMetricStateOperationDescriptor,
            pm.SetComponentStateOperationDescriptor,
            pm.SetAlertStateOperationDescriptor,
        ):
            result.extend(self.descriptions.NODETYPE.get(node_type, []))
        return result

    def select_descriptors(self, *codings: list[Coding | CodedValue | str]) -> list[AbstractDescriptorContainer]:
        """Return all descriptor containers that match a path defined by list of codings.

        Example:
        -------
        [Coding('70041')] returns all containers that have Coding('70041') in its Type
        [Coding('70041'), Coding('69650')] : returns all descriptors with Coding('69650')
                                     and parent descriptor Coding('70041')
        [Coding('70041'), Coding('69650'), Coding('69651')] : returns all descriptors with Coding('69651') and
                                     parent descriptor Coding('69650') and parent's parent descriptor Coding('70041')
        It is not necessary that path starts at the top of a mds, it can start anywhere.
        :param codings: each element can be a string (which is handled as a Coding with DEFAULT_CODING_SYSTEM),
                         a Coding or a CodedValue.

        """
        selected_objects = self.descriptions.objects  # start with all objects
        for counter, coding in enumerate(codings):
            # normalize coding
            if isinstance(coding, str):
                coding = Coding(coding)  # noqa: PLW2901
            if counter > 0:
                # replace selected_objects with all children of selected objects
                all_handles = [o.Handle for o in selected_objects]  # pylint: disable=not-an-iterable
                selected_objects = []
                for handle in all_handles:
                    selected_objects.extend(self.descriptions.parent_handle.get(handle, []))
            # filter current list
            selected_objects = [
                o for o in selected_objects if o.Type is not None and have_matching_codes(o.Type, coding)
            ]
        return selected_objects

    def get_all_descriptors_in_subtree(
        self,
        root_descriptor_container: AbstractDescriptorContainer,
        depth_first: bool = True,
        include_root: bool = True,
    ) -> list[AbstractDescriptorContainer]:
        """Return the tree below descriptor_container as a flat list.

        :param root_descriptor_container: root descriptor
        :param depth_first: determines order of returned list.
               If depth_first=True result has all leaves on top, otherwise at the end.
        :param include_root: if True descriptor_container itself is also part of returned list
        :return: a list of DescriptorContainer objects.
        """
        result = []

        def _getchildren(parent: AbstractDescriptorContainer):
            child_containers = self.descriptions.parent_handle.get(parent.Handle, [])
            if not depth_first:
                result.extend(child_containers)
            apply_map(_getchildren, child_containers)
            if depth_first:
                result.extend(child_containers)

        if include_root and not depth_first:
            result.append(root_descriptor_container)
        _getchildren(root_descriptor_container)
        if include_root and depth_first:
            result.append(root_descriptor_container)
        return result

    def rm_descriptors_and_states(self, descriptor_containers: list[AbstractDescriptorContainer]):
        """Delete descriptors and all related states."""
        deleted_descriptors = {}
        deleted_states = {}
        for descriptor_container in descriptor_containers:
            self._logger.debug(  # noqa: PLE1205
                'rm Descriptor node {} handle {}',
                descriptor_container.NODETYPE,
                descriptor_container.Handle,
            )
            self.descriptions.remove_object(descriptor_container)
            deleted_descriptors[descriptor_container.Handle] = descriptor_container
            for m_key in (self.states, self.context_states):
                state_containers = m_key.descriptor_handle.get(descriptor_container.Handle)
                if state_containers is not None:
                    # make a copy, otherwise remove_objects will manipulate same list in place
                    state_containers = state_containers[:]
                    self._logger.debug(  # noqa: PLE1205
                        'rm {} states(s) associated to descriptor {} ',
                        len(state_containers),
                        descriptor_container.Handle,
                    )
                    m_key.remove_objects(state_containers)
                    deleted_states[descriptor_container.Handle] = state_containers

        if deleted_descriptors:
            self.deleted_descriptors_by_handle = deleted_descriptors
        if deleted_states:
            self.deleted_states_by_handle = deleted_states

    def rm_descriptor_by_handle(self, handle: str):
        """Delete descriptor and the subtree and related states."""
        descriptor_container = self.descriptions.handle.get_one(handle, allow_none=True)
        if descriptor_container is not None:
            all_descriptors = self.get_all_descriptors_in_subtree(descriptor_container)
            self.rm_descriptors_and_states(all_descriptors)

    def get_entity(self, handle: str) -> Entity:
        """Return descriptor and state as Entity."""
        descr = self.descriptions.handle.get_one(handle)
        state = self.states.descriptor_handle.get_one(handle)
        return Entity(self, descr, state)

    def get_context_entity(self, handle: str) -> MultiStateEntity:
        """Return descriptor and states as MultiStateEntity."""
        descr = self.descriptions.handle.get_one(handle)
        states = self.context_states.descriptor_handle.get(handle, [])
        return MultiStateEntity(self, descr, states)

    def has_multiple_mds(self) -> bool:
        """Check if there is more than one mds in mdib (convenience method)."""
        all_mds_descriptors = self.descriptions.NODETYPE.get(self.data_model.pm_names.MdsDescriptor)
        return len(all_mds_descriptors) > 1



"""Implementation of state containers from biceps."""

from __future__ import annotations

import inspect
import sys
import time
import urllib.parse
import uuid
from typing import TYPE_CHECKING, Protocol

from sdc11073.xml_types import ext_qnames as ext
from sdc11073.xml_types import pm_qnames as pm
from sdc11073.xml_types import pm_types
from sdc11073.xml_types import xml_structure as x_struct

from .containerbase import ContainerBase

if TYPE_CHECKING:
    from decimal import Decimal

    from lxml import etree

    from sdc11073 import xml_utils
    from sdc11073.location import SdcLocation
    from sdc11073.namespaces import NamespaceHelper
    from sdc11073.xml_types.isoduration import DurationType
    from sdc11073.xml_types.xml_structure import ExtensionLocalValue

    from .descriptorcontainers import AbstractDescriptorProtocol


class AbstractStateProtocol(Protocol):
    """The common Interface of all states."""

    NODETYPE: etree.QName
    is_state_container: bool
    is_realtime_sample_array_metric_state: bool
    is_metric_state: bool
    is_operational_state: bool
    is_component_state: bool
    is_alert_state: bool
    is_alert_signal: bool
    is_alert_condition: bool
    is_multi_state: bool
    is_context_state: bool

    DescriptorHandle: str
    DescriptorVersion: int
    StateVersion: int

    def __init__(self, descriptor_container: AbstractDescriptorProtocol | None): ...

    def update_from_other_container(self, other: AbstractStateProtocol, skipped_properties: list[str] | None = None):
        """Copy all properties except the skipped ones to self."""

    def update_from_node(self, node: xml_utils.LxmlElement):
        """Update members from node."""


class AbstractStateContainer(ContainerBase):
    """Base class of all states."""

    # these class variables allow easy type-checking. Derived classes will set corresponding values to True
    is_state_container = True
    is_realtime_sample_array_metric_state = False
    is_metric_state = False
    is_operational_state = False
    is_component_state = False
    is_alert_state = False
    is_alert_signal = False
    is_alert_condition = False
    is_multi_state = False
    is_context_state = False

    Extension: ExtensionLocalValue = x_struct.ExtensionNodeProperty(ext.Extension)
    DescriptorHandle: str = x_struct.HandleRefAttributeProperty('DescriptorHandle', is_optional=False)
    DescriptorVersion: int = x_struct.ReferencedVersionAttributeProperty('DescriptorVersion', implied_py_value=0)
    StateVersion: int = x_struct.VersionCounterAttributeProperty('StateVersion', implied_py_value=0)
    _props = ('Extension', 'DescriptorHandle', 'DescriptorVersion', 'StateVersion')

    def __init__(self, descriptor_container: AbstractDescriptorProtocol):
        super().__init__()
        self.descriptor_container = descriptor_container
        if descriptor_container is not None:
            # pylint: disable=invalid-name
            self.DescriptorHandle = descriptor_container.Handle
            self.DescriptorVersion = descriptor_container.DescriptorVersion
            # pylint: enable=invalid-name

    def mk_state_node(
        self,
        tag: etree.QName,
        nsmapper: NamespaceHelper,
        set_xsi_type: bool = True,
    ) -> xml_utils.LxmlElement:
        """Create an etree node from instance data."""
        return super().mk_node(tag, nsmapper, set_xsi_type=set_xsi_type)

    def update_from_other_container(self, other: AbstractStateContainer, skipped_properties: list[str] | None = None):
        """Copy all properties except the skipped ones to self."""
        if other.DescriptorHandle != self.DescriptorHandle:
            msg = (
                f'Update from a node with different descriptor handle is not possible! '
                f'Have "{self.DescriptorHandle}", got "{other.DescriptorHandle}"'
            )
            raise ValueError(msg)
        self._update_from_other(other, skipped_properties)
        self.node = other.node

    def increment_state_version(self):
        """Add one."""
        self.StateVersion += 1

    def update_descriptor_version(self):
        """Set self.DescriptorVersion to version of descriptor."""
        if self.descriptor_container is None:
            msg = f'State {self} has no descriptor_container'
            raise ValueError(msg)
        if self.descriptor_container.DescriptorVersion != self.DescriptorVersion:
            self.DescriptorVersion = self.descriptor_container.DescriptorVersion

    @property
    def source_mds(self) -> str:
        """Get source mds handle."""
        return self.descriptor_container.source_mds

    def __repr__(self) -> str:
        return f'{self.__class__.__name__} DescriptorHandle="{self.DescriptorHandle}" StateVersion={self.StateVersion}'

    @classmethod
    def from_node(
        cls,
        node: xml_utils.LxmlElement,
        descriptor_container: AbstractDescriptorProtocol | None = None,
    ) -> AbstractStateContainer:
        """Create an instance from XML node."""
        obj = cls(descriptor_container)
        obj.update_from_node(node)
        return obj


class AbstractOperationStateContainer(AbstractStateContainer):
    """Represents AbstractOperationState in BICEPS."""

    NODETYPE = pm.AbstractOperationState
    is_operational_state = True
    OperatingMode: pm_types.OperatingMode = x_struct.EnumAttributeProperty(
        'OperatingMode',
        default_py_value=pm_types.OperatingMode.ENABLED,
        enum_cls=pm_types.OperatingMode,
    )
    _props = ('OperatingMode',)


class SetValueOperationStateContainer(AbstractOperationStateContainer):
    """Represents SetValueOperationState in BICEPS."""

    NODETYPE = pm.SetValueOperationState
    AllowedRange: list[pm_types.Range] = x_struct.SubElementListProperty(pm.AllowedRange, value_class=pm_types.Range)
    _props = ('AllowedRange',)


class AllowedValuesType(pm_types.PropertyBasedPMType):  # pylint: disable=invalid-name
    """Represents a list of values, in xml it is a list of pm.Value elements with one value as text."""

    Value: list[str] = x_struct.SubElementStringListProperty(pm.Value)
    _props = ('Value',)

    def is_empty(self) -> bool:
        """Return True if Value is empty."""
        return self.Value is None or len(self.Value) == 0


class SetStringOperationStateContainer(AbstractOperationStateContainer):
    """Represents SetStringOperationState in BICEPS."""

    NODETYPE = pm.SetStringOperationState
    AllowedValues: AllowedValuesType = x_struct.SubElementWithSubElementListProperty(
        pm.AllowedValues,
        default_py_value=AllowedValuesType(),
        value_class=AllowedValuesType,
    )
    _props = ('AllowedValues',)


class ActivateOperationStateContainer(AbstractOperationStateContainer):
    """Represents ActivateOperationState in BICEPS."""

    NODETYPE = pm.ActivateOperationState


class SetContextStateOperationStateContainer(AbstractOperationStateContainer):
    """Represents SetContextStateOperationState in BICEPS."""

    NODETYPE = pm.SetContextStateOperationState


class SetMetricStateOperationStateContainer(AbstractOperationStateContainer):
    """Represents SetMetricStateOperationState in BICEPS."""

    NODETYPE = pm.SetMetricStateOperationState


class SetComponentStateOperationStateContainer(AbstractOperationStateContainer):
    """Represents SetComponentStateOperationState in BICEPS."""

    NODETYPE = pm.SetComponentStateOperationState


class SetAlertStateOperationStateContainer(AbstractOperationStateContainer):
    """Represents SetAlertStateOperationState in BICEPS."""

    NODETYPE = pm.SetAlertStateOperationState


class AbstractMetricStateContainer(AbstractStateContainer):
    """Represents AbstractMetricState in BICEPS."""

    is_metric_state = True
    BodySite: list[pm_types.CodedValue] = x_struct.SubElementListProperty(pm.BodySite, value_class=pm_types.CodedValue)
    PhysicalConnector: pm_types.PhysicalConnectorInfo = x_struct.SubElementProperty(
        pm.PhysicalConnector,
        value_class=pm_types.PhysicalConnectorInfo,
        is_optional=True,
    )
    ActivationState: pm_types.ComponentActivation | None = x_struct.EnumAttributeProperty(
        'ActivationState',
        implied_py_value=pm_types.ComponentActivation.ON,
        enum_cls=pm_types.ComponentActivation,
    )
    ActiveDeterminationPeriod: DurationType | None = x_struct.DurationAttributeProperty('ActiveDeterminationPeriod')
    LifeTimePeriod: DurationType | None = x_struct.DurationAttributeProperty('LifeTimePeriod')
    _props = ('BodySite', 'PhysicalConnector', 'ActivationState', 'ActiveDeterminationPeriod', 'LifeTimePeriod')


class MetricStateProtocol(AbstractStateProtocol):
    """Metric state protocol."""

    MetricValue: None | pm_types.NumericMetricValue | pm_types.StringMetricValue | pm_types.SampleArrayValue

    def mk_metric_value(self) -> pm_types.NumericMetricValue | pm_types.StringMetricValue | pm_types.SampleArrayValue:
        """Create a metric value."""


class NumericMetricStateContainer(AbstractMetricStateContainer):
    """Represents NumericMetricState in BICEPS."""

    NODETYPE = pm.NumericMetricState
    MetricValue: pm_types.NumericMetricValue | None = x_struct.SubElementProperty(
        pm.MetricValue,
        value_class=pm_types.NumericMetricValue,
        is_optional=True,
    )
    PhysiologicalRange: list[pm_types.Range] = x_struct.SubElementListProperty(
        pm.PhysiologicalRange,
        value_class=pm_types.Range,
    )
    ActiveAveragingPeriod: DurationType | None = x_struct.DurationAttributeProperty('ActiveAveragingPeriod')
    _props = ('MetricValue', 'PhysiologicalRange', 'ActiveAveragingPeriod')

    def mk_metric_value(self) -> pm_types.NumericMetricValue:
        """Instantiate self.MetricValue."""
        if self.MetricValue is None:
            self.MetricValue = pm_types.NumericMetricValue()
            return self.MetricValue
        msg = f'State (descriptor handle="{self.DescriptorHandle}") already has a metric value'
        raise ValueError(msg)


class StringMetricStateContainer(AbstractMetricStateContainer):
    """Represents StringMetricState in BICEPS."""

    NODETYPE = pm.StringMetricState
    MetricValue: pm_types.StringMetricValue | None = x_struct.SubElementProperty(
        pm.MetricValue,
        value_class=pm_types.StringMetricValue,
        is_optional=True,
    )
    _props = ('MetricValue',)

    def mk_metric_value(self) -> pm_types.StringMetricValue:
        """Instantiate self.MetricValue."""
        if self.MetricValue is None:
            self.MetricValue = pm_types.StringMetricValue()
            return self.MetricValue
        msg = f'State (descriptor handle="{self.DescriptorHandle}") already has a metric value'
        raise ValueError(msg)


class EnumStringMetricStateContainer(StringMetricStateContainer):
    """Represents EnumStringMetricState in BICEPS."""

    NODETYPE = pm.EnumStringMetricState


class RealTimeSampleArrayMetricStateContainer(AbstractMetricStateContainer):
    """Represents RealTimeSampleArrayMetricState in BICEPS."""

    NODETYPE = pm.RealTimeSampleArrayMetricState
    is_realtime_sample_array_metric_state = True
    MetricValue: pm_types.SampleArrayValue | None = x_struct.SubElementProperty(
        pm.MetricValue,
        value_class=pm_types.SampleArrayValue,
        is_optional=True,
    )
    PhysiologicalRange: list[pm_types.Range] = x_struct.SubElementListProperty(
        pm.PhysiologicalRange,
        value_class=pm_types.Range,
    )
    _props = ('MetricValue', 'PhysiologicalRange')

    def mk_metric_value(self) -> pm_types.SampleArrayValue:
        """Instantiate self.MetricValue."""
        if self.MetricValue is None:
            self.MetricValue = pm_types.SampleArrayValue()
            return self.MetricValue
        msg = f'State (descriptor handle="{self.DescriptorHandle}") already has a metric value'
        raise ValueError(msg)

    def __repr__(self) -> str:
        samples_count = 0
        if self.MetricValue is not None and self.MetricValue.Samples is not None:
            samples_count = len(self.MetricValue.Samples)
        return (
            f'{self.__class__.__name__} DescriptorHandle="{self.DescriptorHandle}" '
            f'Activation="{self.ActivationState}" Samples={samples_count}'
        )


class DistributionSampleArrayMetricStateContainer(AbstractMetricStateContainer):
    """Represents DistributionSampleArrayMetricState in BICEPS."""

    NODETYPE = pm.DistributionSampleArrayMetricState
    MetricValue = x_struct.SubElementProperty(pm.MetricValue, value_class=pm_types.SampleArrayValue, is_optional=True)
    PhysiologicalRange: list[pm_types.Range] = x_struct.SubElementListProperty(
        pm.PhysiologicalRange,
        value_class=pm_types.Range,
    )
    _props = ('MetricValue', 'PhysiologicalRange')

    def mk_metric_value(self) -> pm_types.SampleArrayValue:
        """Instantiate self.MetricValue."""
        if self.MetricValue is None:
            self.MetricValue = pm_types.SampleArrayValue()
            return self.MetricValue
        msg = f'State (descriptor handle="{self.DescriptorHandle}") already has a metric value'
        raise ValueError(msg)

    def __repr__(self) -> str:
        samples_count = 0
        if self.MetricValue is not None and self.MetricValue.Samples is not None:
            samples_count = len(self.MetricValue.Samples)
        return (
            f'{self.__class__.__name__} DescriptorHandle="{self.DescriptorHandle}" '
            f'Activation="{self.ActivationState}" Samples={samples_count}'
        )


class AbstractDeviceComponentStateContainer(AbstractStateContainer):
    """Represents AbstractDeviceComponentState in BICEPS."""

    is_component_state = True
    CalibrationInfo: pm_types.CalibrationInfo | None = x_struct.SubElementProperty(
        pm.CalibrationInfo,
        value_class=pm_types.CalibrationInfo,
        is_optional=True,
    )
    NextCalibration: pm_types.CalibrationInfo | None = x_struct.SubElementProperty(
        pm.NextCalibration,
        value_class=pm_types.CalibrationInfo,
        is_optional=True,
    )
    PhysicalConnector: pm_types.PhysicalConnectorInfo | None = x_struct.SubElementProperty(
        pm.PhysicalConnector,
        value_class=pm_types.PhysicalConnectorInfo,
        is_optional=True,
    )

    ActivationState: pm_types.ComponentActivation = x_struct.EnumAttributeProperty(
        'ActivationState',
        enum_cls=pm_types.ComponentActivation,
        implied_py_value=pm_types.ComponentActivation.ON,
    )
    OperatingHours: int | None = x_struct.UnsignedIntAttributeProperty('OperatingHours')
    OperatingCycles: int | None = x_struct.UnsignedIntAttributeProperty('OperatingCycles')
    _props = (
        'CalibrationInfo',
        'NextCalibration',
        'PhysicalConnector',
        'ActivationState',
        'OperatingHours',
        'OperatingCycles',
    )


class AbstractComplexDeviceComponentStateContainer(AbstractDeviceComponentStateContainer):
    """Represents AbstractComplexDeviceComponentState in BICEPS."""

    NODETYPE = pm.AbstractComplexDeviceComponentState


class MdsStateContainer(AbstractComplexDeviceComponentStateContainer):
    """Represents MdsState in BICEPS."""

    NODETYPE = pm.MdsState
    OperatingJurisdiction: pm_types.OperatingJurisdiction | None = x_struct.SubElementProperty(
        pm.OperatingJurisdiction,
        value_class=pm_types.OperatingJurisdiction,
        is_optional=True,
    )
    Lang: str = x_struct.StringAttributeProperty('Lang', implied_py_value='en')
    OperatingMode: pm_types.MdsOperatingMode = x_struct.EnumAttributeProperty(
        'OperatingMode',
        implied_py_value=pm_types.MdsOperatingMode.NORMAL,
        enum_cls=pm_types.MdsOperatingMode,
    )
    _props = ('OperatingJurisdiction', 'OperatingMode', 'Lang')


class ScoStateContainer(AbstractDeviceComponentStateContainer):
    """Represents ScoState in BICEPS."""

    NODETYPE = pm.ScoState
    OperationGroup: list[pm_types.OperationGroup] = x_struct.SubElementListProperty(
        pm.OperationGroup,
        value_class=pm_types.OperationGroup,
    )
    InvocationRequested: list[str] | None = x_struct.OperationRefListAttributeProperty('InvocationRequested')
    InvocationRequired: list[str] | None = x_struct.OperationRefListAttributeProperty('InvocationRequired')
    _props = ('OperationGroup', 'InvocationRequested', 'InvocationRequired')


class VmdStateContainer(AbstractComplexDeviceComponentStateContainer):
    """Represents VmdState in BICEPS."""

    NODETYPE = pm.VmdState
    OperatingJurisdiction: pm_types.OperatingJurisdiction = x_struct.SubElementProperty(
        pm.OperatingJurisdiction,
        value_class=pm_types.OperatingJurisdiction,
        is_optional=True,
    )
    _props = ('OperatingJurisdiction',)


class ChannelStateContainer(AbstractDeviceComponentStateContainer):
    """Represents ChannelState in BICEPS."""

    NODETYPE = pm.ChannelState


class ClockStateContainer(AbstractDeviceComponentStateContainer):
    """Represents ClockState in BICEPS."""

    NODETYPE = pm.ClockState
    ActiveSyncProtocol: pm_types.CodedValue | None = x_struct.SubElementProperty(
        pm.ActiveSyncProtocol,
        value_class=pm_types.CodedValue,
        is_optional=True,
    )
    ReferenceSource: list[str] = x_struct.SubElementStringListProperty(pm.ReferenceSource)
    DateAndTime: float = x_struct.CurrentTimestampAttributeProperty('DateAndTime')
    RemoteSync: bool = x_struct.BooleanAttributeProperty('RemoteSync', default_py_value=True, is_optional=False)
    Accuracy: Decimal | None = x_struct.DecimalAttributeProperty('Accuracy')
    LastSet: float | None = x_struct.TimestampAttributeProperty('LastSet')
    TimeZone: str | None = x_struct.TimeZoneAttributeProperty('TimeZone')
    CriticalUse: bool = x_struct.BooleanAttributeProperty('CriticalUse', implied_py_value=False)
    _props = (
        'ActiveSyncProtocol',
        'ReferenceSource',
        'DateAndTime',
        'RemoteSync',
        'Accuracy',
        'LastSet',
        'TimeZone',
        'CriticalUse',
    )


class SystemContextStateContainer(AbstractDeviceComponentStateContainer):
    """Represents SystemContextState in BICEPS."""

    NODETYPE = pm.SystemContextState


class BatteryStateContainer(AbstractDeviceComponentStateContainer):
    """Represents BatteryState in BICEPS."""

    class ChargeStatusEnum(pm_types.StringEnum):
        """ChargeStatusEnum contains the allowed values for ChargeStatus."""

        FULL = 'Ful'
        CHARGING = 'ChB'
        DISCHARGING = 'DisChB'
        EMPTY = 'DEB'

    NODETYPE = pm.BatteryState
    CapacityRemaining: pm_types.Measurement | None = x_struct.SubElementProperty(
        pm.CapacityRemaining,
        value_class=pm_types.Measurement,
        is_optional=True,
    )
    Voltage: pm_types.Measurement | None = x_struct.SubElementProperty(
        pm.Voltage,
        value_class=pm_types.Measurement,
        is_optional=True,
    )
    Current: pm_types.Measurement | None = x_struct.SubElementProperty(
        pm.Current,
        value_class=pm_types.Measurement,
        is_optional=True,
    )
    Temperature: pm_types.Measurement | None = x_struct.SubElementProperty(
        pm.Temperature,
        value_class=pm_types.Measurement,
        is_optional=True,
    )
    RemainingBatteryTime: pm_types.Measurement | None = x_struct.SubElementProperty(
        pm.RemainingBatteryTime,
        value_class=pm_types.Measurement,
        is_optional=True,
    )
    ChargeStatus: ChargeStatusEnum | None = x_struct.EnumAttributeProperty('ChargeStatus', enum_cls=ChargeStatusEnum)
    ChargeCycles: int | None = x_struct.UnsignedIntAttributeProperty(
        'ChargeCycles',
    )  # Number of charge/discharge cycles.
    _props = (
        'CapacityRemaining',
        'Voltage',
        'Current',
        'Temperature',
        'RemainingBatteryTime',
        'ChargeStatus',
        'ChargeCycles',
    )


class AbstractAlertStateContainer(AbstractStateContainer):
    """Represents AbstractAlertState in BICEPS."""

    is_alert_state = True
    ActivationState: pm_types.AlertActivation = x_struct.EnumAttributeProperty(
        'ActivationState',
        default_py_value=pm_types.AlertActivation.ON,
        enum_cls=pm_types.AlertActivation,
        is_optional=False,
    )
    _props = ('ActivationState',)


class AlertSystemStateContainer(AbstractAlertStateContainer):
    """Represents AlertSystemState in BICEPS."""

    NODETYPE = pm.AlertSystemState
    SystemSignalActivation: list[pm_types.SystemSignalActivation] = x_struct.SubElementListProperty(
        pm.SystemSignalActivation,
        value_class=pm_types.SystemSignalActivation,
    )
    LastSelfCheck: float | None = x_struct.TimestampAttributeProperty('LastSelfCheck')
    SelfCheckCount: int | None = x_struct.IntegerAttributeProperty('SelfCheckCount')
    PresentPhysiologicalAlarmConditions: list[str] | None = x_struct.AlertConditionRefListAttributeProperty(
        'PresentPhysiologicalAlarmConditions',
    )
    PresentTechnicalAlarmConditions: list[str] | None = x_struct.AlertConditionRefListAttributeProperty(
        'PresentTechnicalAlarmConditions',
    )
    _props = (
        'SystemSignalActivation',
        'LastSelfCheck',
        'SelfCheckCount',
        'PresentPhysiologicalAlarmConditions',
        'PresentTechnicalAlarmConditions',
    )

    def __repr__(self) -> str:
        return (
            f'{self.__class__.__name__} DescriptorHandle="{self.DescriptorHandle}" '
            f'StateVersion={self.StateVersion} LastSelfCheck={self.LastSelfCheck} '
            f'SelfCheckCount={self.SelfCheckCount} Activation={self.ActivationState}'
        )


class AlertSignalStateContainer(AbstractAlertStateContainer):
    """Represents AlertSignalState in BICEPS."""

    is_alert_signal = True
    NODETYPE = pm.AlertSignalState
    ActualSignalGenerationDelay: DurationType | None = x_struct.DurationAttributeProperty('ActualSignalGenerationDelay')
    Presence: pm_types.AlertSignalPresence = x_struct.EnumAttributeProperty(
        'Presence',
        implied_py_value=pm_types.AlertSignalPresence.OFF,
        enum_cls=pm_types.AlertSignalPresence,
    )
    Location: pm_types.AlertSignalPrimaryLocation = x_struct.EnumAttributeProperty(
        'Location',
        implied_py_value=pm_types.AlertSignalPrimaryLocation.LOCAL,
        enum_cls=pm_types.AlertSignalPrimaryLocation,
    )
    Slot: int | None = x_struct.UnsignedIntAttributeProperty('Slot')
    _props = ('ActualSignalGenerationDelay', 'Presence', 'Location', 'Slot')

    def __init__(self, descriptor_container: AbstractDescriptorProtocol):
        super().__init__(descriptor_container)
        self.last_updated = time.time()

    def __repr__(self) -> str:
        return (
            f'{self.__class__.__name__} DescriptorHandle="{self.DescriptorHandle}" '
            f'StateVersion={self.StateVersion} Location={self.Location} '
            f'Activation={self.ActivationState} Presence={self.Presence}'
        )


class AlertConditionStateContainer(AbstractAlertStateContainer):
    """Represents AlertConditionState in BICEPS."""

    is_alert_condition = True
    NODETYPE = pm.AlertConditionState
    ActualConditionGenerationDelay: DurationType | None = x_struct.DurationAttributeProperty(
        'ActualConditionGenerationDelay',
    )
    ActualPriority: pm_types.AlertConditionPriority | None = x_struct.EnumAttributeProperty(
        'ActualPriority',
        enum_cls=pm_types.AlertConditionPriority,
    )
    Rank: int | None = x_struct.IntegerAttributeProperty('Rank')
    DeterminationTime: float | None = x_struct.TimestampAttributeProperty('DeterminationTime')
    Presence: bool = x_struct.BooleanAttributeProperty('Presence', implied_py_value=False)
    _props = ('ActualConditionGenerationDelay', 'ActualPriority', 'Rank', 'DeterminationTime', 'Presence')

    def __repr__(self) -> str:
        return (
            f'{self.__class__.__name__} DescriptorHandle="{self.DescriptorHandle}" '
            f'StateVersion={self.StateVersion} '
            f'Activation={self.ActivationState} Presence={self.Presence}'
        )


class LimitAlertConditionStateContainer(AlertConditionStateContainer):
    """Represents LimitAlertConditionState in BICEPS."""

    NODETYPE = pm.LimitAlertConditionState
    Limits: pm_types.Range = x_struct.SubElementProperty(
        pm.Limits,
        value_class=pm_types.Range,
        default_py_value=pm_types.Range(),
    )
    MonitoredAlertLimits: pm_types.AlertConditionMonitoredLimits = x_struct.EnumAttributeProperty(
        'MonitoredAlertLimits',
        default_py_value=pm_types.AlertConditionMonitoredLimits.NONE,
        enum_cls=pm_types.AlertConditionMonitoredLimits,
        is_optional=False,
    )
    AutoLimitActivationState: pm_types.AlertActivation | None = x_struct.EnumAttributeProperty(
        'AutoLimitActivationState',
        enum_cls=pm_types.AlertActivation,
    )
    _props = ('Limits', 'MonitoredAlertLimits', 'AutoLimitActivationState')


class AbstractMultiStateProtocol(AbstractStateProtocol):
    """Protocol for AbstractContextStateContainer."""

    Handle: str


class AbstractMultiStateContainer(AbstractStateContainer):
    """Represents AbstractMultiState in BICEPS."""

    is_multi_state = True
    Category: pm_types.CodedValue | None = x_struct.SubElementProperty(
        pm.Category,
        value_class=pm_types.CodedValue,
        is_optional=True,
    )
    Handle: str = x_struct.HandleAttributeProperty('Handle', is_optional=False)
    _props = ('Category', 'Handle')

    def __init__(self, descriptor_container: AbstractDescriptorProtocol, handle: str | None = None):
        super().__init__(descriptor_container)
        self.Handle = handle  # pylint: disable=invalid-name

    def update_from_other_container(
        self,
        other: AbstractMultiStateContainer,
        skipped_properties: list[str] | None = None,
    ):
        """Copy all properties except the skipped ones to self.

        Accept node only if DescriptorHandle and Handle match.
        """
        if self.Handle is not None and other.Handle != self.Handle:
            msg = (
                f'Update from a node with different handle is not possible! Have "{self.Handle}", got "{other.Handle}"'
            )
            raise ValueError(
                msg,
            )
        super().update_from_other_container(other, skipped_properties)

    def mk_state_node(
        self,
        tag: etree.QName,
        nsmapper: NamespaceHelper,
        set_xsi_type: bool = True,
    ) -> xml_utils.LxmlElement:
        """Create an etree node from instance data."""
        if self.Handle is None:
            self.Handle = uuid.uuid4().hex
        return super().mk_state_node(tag, nsmapper, set_xsi_type)

    def __repr__(self) -> str:
        return (
            f'{self.__class__.__name__} DescriptorHandle="{self.DescriptorHandle}" '
            f'Handle="{self.Handle}" type={self.NODETYPE}'
        )


class AbstractContextStateContainer(AbstractMultiStateContainer):
    """Represents AbstractContextState in BICEPS."""

    is_context_state = True
    Validator: list[pm_types.InstanceIdentifier] = x_struct.SubElementListProperty(
        pm.Validator,
        value_class=pm_types.InstanceIdentifier,
    )
    Identification: list[pm_types.InstanceIdentifier] = x_struct.SubElementListProperty(
        pm.Identification,
        value_class=pm_types.InstanceIdentifier,
    )
    ContextAssociation: pm_types.ContextAssociation = x_struct.EnumAttributeProperty(
        'ContextAssociation',
        enum_cls=pm_types.ContextAssociation,
        implied_py_value=pm_types.ContextAssociation.NO_ASSOCIATION,
    )
    BindingMdibVersion: int | None = x_struct.ReferencedVersionAttributeProperty('BindingMdibVersion')
    UnbindingMdibVersion: int | None = x_struct.ReferencedVersionAttributeProperty('UnbindingMdibVersion')
    BindingStartTime: float | None = x_struct.TimestampAttributeProperty('BindingStartTime')
    BindingEndTime: float | None = x_struct.TimestampAttributeProperty('BindingEndTime')
    _props = (
        'Validator',
        'Identification',
        'ContextAssociation',
        'BindingMdibVersion',
        'UnbindingMdibVersion',
        'BindingStartTime',
        'BindingEndTime',
    )


class LocationContextStateContainer(AbstractContextStateContainer):
    """Represents LocationContextState in BICEPS."""

    NODETYPE = pm.LocationContextState
    # Technically LocationDetail does not need a default_py_value because it is optional,
    # but it makes it much easier for users if LocationDetail always exists.
    # The consequence is a difference in generated xml: The LocationDetail node will always exist, even if nothing was set.  # noqa: E501, W505
    # To circumvent this you can set LocationDetail to None.
    LocationDetail: pm_types.LocationDetail = x_struct.SubElementProperty(
        pm.LocationDetail,
        value_class=pm_types.LocationDetail,
        default_py_value=pm_types.LocationDetail(),
        is_optional=True,
    )
    _props = ('LocationDetail',)

    def update_from_sdc_location(self, sdc_location: SdcLocation):
        """Set members according to sdc_location using the fallback instance identifier algorithm from GLUE 9.4.1.1."""
        if self.LocationDetail is None:
            self.LocationDetail = pm_types.LocationDetail()
        self.LocationDetail.PoC = sdc_location.poc
        self.LocationDetail.Room = sdc_location.rm
        self.LocationDetail.Bed = sdc_location.bed
        self.LocationDetail.Facility = sdc_location.fac
        self.LocationDetail.Building = sdc_location.bldng
        self.LocationDetail.Floor = sdc_location.flr
        self.ContextAssociation = pm_types.ContextAssociation.ASSOCIATED

        extension_string = self._loc_extension_segment(sdc_location)
        self.Identification = [
            pm_types.InstanceIdentifier(root='sdc.ctxt.loc.detail', extension_string=extension_string),
        ]

    def _loc_extension_segment(self, sdc_location: SdcLocation) -> str:
        """Build location extension segment as defined in GLUE 9.4.1.1 Fallback instance identifier algorithm.

        None elements are replaces by an empty string.
        """
        extension = '/'.join(
            [urllib.parse.quote(getattr(sdc_location, attr) or '', safe='') for attr in sdc_location.url_elements],
        )
        if extension == '/////':  # extension segment requires at least 1 entry with 1 character
            msg = 'Location extension segment is empty, at least one element must be set'
            raise ValueError(msg)
        return extension

    @classmethod
    def from_sdc_location(
        cls,
        descriptor_container: AbstractDescriptorProtocol,
        handle: str,
        sdc_location: SdcLocation,
    ) -> LocationContextStateContainer:
        """Construct LocationContextStateContainer from a sdc location."""
        obj = cls(descriptor_container)
        obj.Handle = handle
        obj.update_from_sdc_location(sdc_location)
        return obj


class PatientContextStateContainer(AbstractContextStateContainer):
    """Represents PatientContextState in BICEPS."""

    NODETYPE = pm.PatientContextState
    # Technically CoreData does not need a default_py_value because it is optional,
    # but it makes it much easier for users if CoreData always exists.
    # The consequence is a difference in generated xml: The CoreData node will always exist, even if nothing was set.
    # To circumvent this you can set CoreData to None.
    CoreData: pm_types.PatientDemographicsCoreData | None = x_struct.SubElementProperty(
        pm.CoreData,
        value_class=pm_types.PatientDemographicsCoreData,
        default_py_value=pm_types.PatientDemographicsCoreData(),
        is_optional=True,
    )
    _props = ('CoreData',)


class WorkflowContextStateContainer(AbstractContextStateContainer):
    """Represents WorkflowContextState in BICEPS."""

    NODETYPE = pm.WorkflowContextState
    WorkflowDetail: pm_types.WorkflowDetail | None = x_struct.SubElementProperty(
        pm.WorkflowDetail,
        value_class=pm_types.WorkflowDetail,
        is_optional=True,
    )
    _props = ('WorkflowDetail',)


class OperatorContextStateContainer(AbstractContextStateContainer):
    """Represents OperatorContextState in BICEPS."""

    NODETYPE = pm.OperatorContextState
    OperatorDetails: pm_types.BaseDemographics | None = x_struct.SubElementProperty(
        pm.OperatorDetails,
        value_class=pm_types.BaseDemographics,
        is_optional=True,
    )
    _props = ('OperatorDetails',)


class MeansContextStateContainer(AbstractContextStateContainer):
    """Represents MeansContextState in BICEPS."""

    NODETYPE = pm.MeansContextState
    # class has no own members


class EnsembleContextStateContainer(AbstractContextStateContainer):
    """Represents EnsembleContextState in BICEPS."""

    NODETYPE = pm.EnsembleContextState
    # class has no own members


# mapping of states: xsi:type information to classes
# find all classes in this module that have a member "NODETYPE"
classes = inspect.getmembers(
    sys.modules[__name__],
    lambda member: inspect.isclass(member) and member.__module__ == __name__,
)
classes_with_nodetype = [c[1] for c in classes if hasattr(c[1], 'NODETYPE') and c[1].NODETYPE is not None]
# make a dictionary from found classes: (Key is NODETYPE, value is the class itself
_state_lookup_by_type = {c.NODETYPE: c for c in classes_with_nodetype}


def get_container_class(type_qname: etree.QName) -> type[AbstractStateProtocol]:
    """Return class for given type.

    :param type_qname: the QName of the expected NODETYPE.
    """
    return _state_lookup_by_type.get(type_qname)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from sdc11073.pysoap.msgfactory import MessageFactory
from sdc11073.pysoap.msgreader import MessageReader
from sdc11073.pysoap.soapclient import SoapClient

from .operations import OperationsManager, OperationsManagerProtocol
from .request_handler_deferred import DispatchKeyRegistryDeferred
from .serviceclients.containmenttreeservice import CTreeServiceClient
from .serviceclients.contextservice import ContextServiceClient
from .serviceclients.descriptioneventservice import DescriptionEventClient
from .serviceclients.getservice import GetServiceClient
from .serviceclients.localizationservice import LocalizationServiceClient
from .serviceclients.setservice import SetServiceClient
from .serviceclients.stateeventservice import StateEventClient
from .serviceclients.waveformservice import WaveformClient
from .subscription import ConsumerSubscriptionManager, ConsumerSubscriptionManagerProtocol

if TYPE_CHECKING:
    from sdc11073.pysoap.soapclient import SoapClientProtocol
    from sdc11073.dispatch.dispatchkey import RequestDispatcherProtocol
    from sdc11073.namespaces import PrefixNamespace


@dataclass()
class SdcConsumerComponents:
    """Dependency injection: This class defines which component implementations the sdc consumer will use."""

    soap_client_class: type[SoapClientProtocol] = None
    msg_factory_class: type[MessageFactory] = None
    msg_reader_class: type[MessageReader] = None
    action_dispatcher_class: type[RequestDispatcherProtocol] = None
    subscription_manager_class: type[ConsumerSubscriptionManagerProtocol] = None
    operations_manager_class: type[OperationsManagerProtocol] | None = None
    service_handlers: list = None
    additional_schema_specs: list[PrefixNamespace] = field(default_factory=list)

    def merge(self, other: SdcConsumerComponents):
        """Add data from other to self."""
        def _merge(attr_name: str):
            other_value = getattr(other, attr_name)
            if other_value:
                setattr(self, attr_name, other_value)

        _merge('msg_factory_class')
        _merge('msg_reader_class')
        _merge('action_dispatcher_class')
        _merge('subscription_manager_class')
        _merge('operations_manager_class')
        if other.service_handlers:
            # append handlers that are not yet present
            for handler in other.service_handlers:
                if handler not in self.service_handlers:
                    self.service_handlers.append(handler)
        self.additional_schema_specs = list(set(self.additional_schema_specs).union(set(other.additional_schema_specs)))


default_sdc_consumer_components = SdcConsumerComponents(
    soap_client_class=SoapClient,
    msg_factory_class=MessageFactory,
    msg_reader_class=MessageReader,
    action_dispatcher_class=DispatchKeyRegistryDeferred,  # defaults to deferred handling
    subscription_manager_class=ConsumerSubscriptionManager,
    operations_manager_class=OperationsManager,
    service_handlers=[CTreeServiceClient,
                      GetServiceClient,
                      StateEventClient,
                      ContextServiceClient,
                      WaveformClient,
                      SetServiceClient,
                      DescriptionEventClient,
                      LocalizationServiceClient,
                      ],
)




class ApiUsageError(Exception):
    """This Exception is thrown when a call is made when it should not be called, e.g. call initialize() twice."""


class HTTPRequestHandlingError(Exception):
    """ This class is used to communicate errors from http request handlers back to http server."""

    def __init__(self, status, reason, soap_fault):
        """
        :param status: integer, e.g. 404
        :param reason: the provided human readable text
        :param soap_fault: soapenvelope.SoapFault instance
        """
        super().__init__()
        self.status = status
        self.reason = reason
        self.soap_fault = soap_fault

    def __repr__(self):
        if self.soap_fault:
            return f'{self.__class__.__name__}(status={self.status}, reason={self.soap_fault})'
        return f'{self.__class__.__name__}(status={self.status}, reason={self.reason})'


class FunctionNotImplementedError(HTTPRequestHandlingError):
    def __init__(self, soap_fault):
        """
        :param soap_fault: soapenvelope.SoapFault instance
        """
        super().__init__(500, 'not implemented', soap_fault)


class InvalidActionError(HTTPRequestHandlingError):
    def __init__(self, soap_fault):
        """
        :param soap_fault: soapenvelope.SoapFault instance
        """
        super().__init__(400, 'Bad Request', soap_fault)


class InvalidPathError(HTTPRequestHandlingError):
    def __init__(self, reason, soap_fault):
        """
        :param soap_fault: soapenvelope.SoapFault instance
        """
        super().__init__(404, reason, soap_fault)

class ValidationError(HTTPRequestHandlingError):
    def __init__(self, reason, soap_fault):
        """
        :param soap_fault: soapenvelope.SoapFault instance
        """
        super().__init__(400, reason, soap_fault)
